'''
**__init__.py: Locate and load the compiled blond library.**

:Authors: Konstantinos Iliakis**
'''

def test():
    """A simple test to verify that blond has been correctly installed.
    """
    try:
        # Test that all the blond modules exist
        # Verify all the dependencies are also installed
        import __future__
        from packaging import version
        import platform
        import h5py
        import matplotlib
        import mpmath
        import numpy
        import scipy

        assert version.parse(matplotlib.__version__) >= version.parse('3.7'), f'matplotlib version {matplotlib.__version__} does not meet minimum required version 3.7'
        assert version.parse(platform.python_version()) >= version.parse('3.8'), f'python version {platform.python_version()} does not meet minimum required version 3.8'
        assert version.parse(numpy.__version__) >= version.parse('1.20'), 'numpy version {numpy.__version__} does not meet minimum required version 1.20'

        from blond import (beam, impedances, input_parameters, llrf, monitors,
                           plots, synchrotron_radiation, toolbox, trackers,
                           utils)
        # This should report the backend that is used
        from blond.utils import bmath
        bmath.report_backend()
        print('\nBLonD installed successfully!')
        return 0
    except Exception as exception:
        print('\nBLonD installation verification failed: ', exception)
        return -1




# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

"""
**Parent class to develop cavity feedback models and various cavity loops for the CERN machines**

:Authors: **Birk Emil Karlsen-Bæck**, **Helga Timko**
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import logging
import sys

from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import numpy as np
import numpy.random as rnd
import scipy.signal
from scipy.interpolate import interp1d

from blond.llrf.impulse_response import (
    SPS3Section200MHzTWC,
    SPS4Section200MHzTWC,
    SPS5Section200MHzTWC,
    cavity_response_sparse_matrix,
)
from blond.llrf.signal_processing import (
    cartesian_to_polar,
    comb_filter,
    feedforward_filter_TWC3,
    feedforward_filter_TWC4,
    feedforward_filter_TWC5,
    modulator,
    moving_average,
    get_power_gen_i,
    polar_to_cartesian,
    rf_beam_current,
    fir_filter_lhc_otfb_coeff,
    smooth_step,
)
from blond.utils import bmath as bm

if TYPE_CHECKING:
    from blond.input_parameters.rf_parameters import RFStation
    from blond.beam.profile import Profile
    from typing import Any


class CavityFeedback(ABC):
    r"""Parent class for implementing cavity feedback models interfacing with BLonD

    Parameters
    ----------
    RFStation : class
        An RFStation type class
    Beam : class
        A Beam type class
    Profile : class
        A Profile type class
    n_cavities : int
        Number of cavities the feedback is acting with
    n_s : int
        The number of RF periods the coarse grid sampling period corresponds to
    n_h : int
        The index of the harmoninc in the RFStation that the feedback should act on
    """

    def __init__(
        self,
        RFStation: RFStation,
        Profile: Profile,
        n_cavities: int,
        n_s: int,
        n_h: int,
        lpf: bool = False,
    ):
        # BLonD classes the feedback should get information from
        self.rfstation = RFStation
        self.profile = Profile
        self.counter = self.rfstation.counter[0]

        # Number of cavities the feedback is working on
        self.n_cavities = int(n_cavities)
        if self.n_cavities < 1:
            raise RuntimeError(
                "ERROR in CavityFeedback: argument" " n_cavities has invalid value!"
            )

        # Apply a low-pass filter to the RF beam current
        self.lpf = lpf

        # The harmonic index the cavity feedback is working on
        self.n_h = int(n_h)
        if self.n_h > self.rfstation.n_rf - 1:
            raise RuntimeError(
                "ERROR in CavityFeedback: argument"
                " n_h is greater than the number of n_rf in RFStation"
            )

        # Sampling time in the model and the number of samples per turn
        self.n_s = int(n_s)
        self.T_s = (
            self.n_s
            * 2
            * np.pi
            / self.rfstation.omega_rf[self.n_h, self.rfstation.counter[0]]
        )
        self.n_coarse = round(self.rfstation.t_rev[0] / self.T_s)
        self.omega_carrier = (
            self.rfstation.omega_rf[self.n_h, self.rfstation.counter[0]] / self.n_s
        )
        self.omega_rf = self.rfstation.omega_rf[self.n_h, self.rfstation.counter[0]]
        self.dT = 0

        # The least amount of arrays needed to feedback to the tracker object
        self.rf_centers = (
            np.arange(self.n_coarse) * self.T_s
            + 0.5 * self.rfstation.t_rf[self.n_h, self.rfstation.counter[0]]
        )
        self.V_SET = np.zeros(2 * self.n_coarse, dtype=complex)
        self.I_BEAM_COARSE = np.zeros(2 * self.n_coarse, dtype=complex)
        self.I_BEAM_FINE = np.zeros(self.profile.n_slices, dtype=complex)
        self.V_ANT_COARSE = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_ANT_FINE = np.zeros(self.profile.n_slices, dtype=complex)
        self.I_GEN_COARSE = np.zeros(2 * self.n_coarse, dtype=complex)
        self.I_GEN_FINE = np.zeros(self.profile.n_slices, dtype=complex)

    @abstractmethod
    def circuit_track(self, no_beam: bool = False):
        r"""Method to track circuit of the feedback. This is meant to be implemented in the child class by the user.
        The only requirement for this method is that it has to update the V_ANT_FINE and V_SET arrays turn-by-turn.
        """
        pass

    @abstractmethod
    def update_fb_variables(self):
        r"""Method to update the variables specific to the feedback.
        This is meant to be implemented in the child class by the user."""
        pass

    def track_no_beam(self, n_pretrack: int = None):
        r"""Tracking method of the cavity feedback without beam in the accelerator"""

        self.update_rf_variables()
        self.update_fb_variables()
        if n_pretrack is None:
            self.circuit_track(no_beam=True)
        else:
            for i in range(n_pretrack):
                self.circuit_track(no_beam=True)

    def track(self):
        r"""Tracking method of the cavity feedback"""
        # Update parameters from rest of BLonD classes
        self.update_rf_variables()
        self.update_fb_variables()

        # Get rf beam current
        self.rf_beam_current(lpf=self.lpf)

        # Tracking circuit model of feedback
        self.circuit_track()

        # Convert to amplitude and phase
        self.V_corr, self.alpha_sum = cartesian_to_polar(
            self.V_ANT_FINE[-self.profile.n_slices :]
        )

        # Calculate OTFB correction w.r.t. RF voltage and phase in RFStation
        self.V_corr /= self.rfstation.voltage[self.n_h, self.rfstation.counter[0]]
        self.phi_corr = self.alpha_sum - np.angle(
            np.interp(
                self.profile.bin_centers, self.rf_centers, self.V_SET[-self.n_coarse :]
            )
        )

    def rf_beam_current(self, lpf: bool = False):
        r"""Calculate RF beam current from beam profile"""

        # Beam current from profile
        self.I_BEAM_COARSE[: self.n_coarse] = self.I_BEAM_COARSE[-self.n_coarse :]
        self.I_BEAM_FINE, self.I_BEAM_COARSE[-self.n_coarse :] = rf_beam_current(
            self.profile,
            self.omega_rf,
            self.rfstation.t_rev[self.counter],
            lpf=lpf,
            downsample={"Ts": self.T_s, "points": self.n_coarse},
            external_reference=True,
            dT=self.dT,
        )

        # Convert RF beam currents to be in units of Amperes
        self.I_BEAM_FINE = self.I_BEAM_FINE / self.profile.bin_size
        self.I_BEAM_COARSE[-self.n_coarse :] = (
            self.I_BEAM_COARSE[-self.n_coarse :] / self.T_s
        )

    def set_point_from_rfstation(self):
        r"""Computes the setpoint in I/Q based on the RF voltage in the RFStation"""

        V_set = polar_to_cartesian(
            self.rfstation.voltage[self.n_h, self.counter] / self.n_cavities, 0
        )

        return V_set * np.ones(self.n_coarse)

    def update_rf_variables(self):
        r"""Updating variables from the other BLonD classes"""

        # Present time step
        self.counter = self.rfstation.counter[0]

        # Present RF angular frequency
        self.omega_rf = self.rfstation.omega_rf[self.n_h, self.counter]

        # Present carrier frequency: main RF frequency
        self.omega_carrier_prev = self.omega_carrier
        self.omega_carrier = self.omega_rf / self.n_s

        # Present sampling time
        self.T_s_prev = self.T_s
        self.T_s = self.n_s * 2 * np.pi / self.omega_rf

        # Update the coarse grid sampling
        self.n_coarse = round(
            self.rfstation.t_rev[self.rfstation.counter[0]] / self.T_s
        )

        # Present coarse grid and save previous turn coarse grid
        self.rf_centers_prev = np.copy(self.rf_centers)

        # Residual part of last turn entering the current turn due to non-integer harmonic number
        self.dT = -self.rfstation.phi_rf[self.n_h, self.counter] / self.omega_rf
        self.rf_centers = (
            np.arange(self.n_coarse) + 0.5 / self.n_s
        ) * self.T_s + self.dT


class SPSCavityLoopCommissioning:
    r"""Class containing commissioning settings for the cavity feedback

    Parameters
    ----------
    debug : bool
        Debugging output active (True/False); default is False
    open_loop : int(bool)
        Open (True) or closed (False) cavity loop; default is False
    open_fb : int(bool)
        Open (True) or closed (False) feedback; default is False
    open_drive : int(bool)
        Open (True) or closed (False) drive; default is False
    open_ff : int(bool)
        Open (True) or closed (False) feed-forward; default is True.
    V_SET : complex array
        Array set point voltage; default is False
    cpp_conv : bool
        Enable (True) or disable (False) convolutions using a C++ implementation; default is False
    pwr_clamp : bool
        Enable (True) or disable (False) power clamping; default is False
    rot_IQ : complex
        Option to rotate the set point and beam induced voltages in the complex plane.
    excitation : bool
        Excite the model with white noise to perform BBNA measurements
    """

    def __init__(
        self,
        debug: bool = False,
        open_loop: bool = False,
        open_fb: bool = False,
        open_drive: bool = False,
        open_ff: bool = True,
        v_set=None,
        cpp_conv: bool = False,
        pwr_clamp: bool = False,
        rot_iq: complex = 1,
        excitation: bool = False,
    ):
        self.debug = bool(debug)
        self.open_loop = 0 if open_loop else 1
        self.open_fb = 0 if open_fb else 1
        self.open_drive = 0 if open_drive else 1
        self.open_ff = 0 if open_ff else 1
        self.V_SET = v_set
        self.cpp_conv = cpp_conv
        self.pwr_clamp = pwr_clamp
        self.rot_iq = rot_iq
        self.excitation = int(excitation)


class LHCCavityLoopCommissioning:
    r"""RF Feedback settings for LHC ACS cavity loop.

    Parameters
    ----------
    alpha : float
        One-turn feedback memory parameter; default is 15/16
    d_phi_ad : float
        Phase misalignment of digital FB w.r.t. analog FB [deg]
    G_a : float
        Analog FB gain [1]
    G_d : float
        Digital FB gain, w.r.t. analog gain [1]
    G_o : float
        One-turn feedback gain
    tau_a : float
        Analog FB delay time [s]
    tau_d : float
        Digital FB delay time [s]
    tau_o : float
        AC-coupling delay time of one-turn feedback [s]
    mu : float
        Coefficient for the tuner algorithm determining time scale; default is -0.0001
    power_thres : float
        Available RF power in the klystron; default is 300 kW
    open_drive : bool
        Open (True) or closed (False) cavity loop at drive; default is False
    open_loop : bool
        Open (True) or closed (False) cavity loop at RFFB; default is False
    open_otfb : bool
        Open (true) or closed (False) one-turn feedback; default is False
    open_rffb : bool
        Open (True) or closed (False) RFFB; default is False
    open_tuner : bool
        Open (True) or closed (False) tuner control; default is False
    clamping : bool
        Simulate clamping (True) or not (False); default is False
    excitation : bool
        Perform BBNA measurement of the feedback (True); default is False
    """

    def __init__(
        self,
        alpha: float = 15 / 16,
        d_phi_ad: float = 0,
        G_a: float = 0.00001,
        G_d: float = 10,
        G_o: float = 10,
        tau_a: float = 170e-6,
        tau_d: float = 400e-6,
        tau_o: float = 110e-6,
        mu: float = -0.0001,
        power_thres: float = 300e3,
        open_drive: bool = False,
        open_loop: bool = False,
        open_otfb: bool = False,
        open_rffb: bool = False,
        open_tuner: bool = False,
        clamping: bool = False,
        excitation: bool = False,
        excitation_otfb_1: bool = False,
        excitation_otfb_2: bool = False,
        seed1: Any = 1234,
        seed2: Any = 7564,
    ):
        # Import variables
        self.alpha = alpha
        self.d_phi_ad = d_phi_ad * np.pi / 180
        self.G_a = G_a
        self.G_d = G_d
        self.G_o = G_o
        self.tau_a = tau_a
        self.tau_d = tau_d
        self.tau_o = tau_o
        self.mu = mu
        self.power_thres = power_thres
        self.excitation = excitation
        self.excitation_otfb_1 = excitation_otfb_1
        self.excitation_otfb_2 = excitation_otfb_2
        self.seed1 = seed1
        self.seed2 = seed2

        # Multiply with zeros if open == True
        self.open_drive = 0 if open_drive else 1
        self.open_drive_inv = 0 if self.open_drive else 1
        self.open_loop = 0 if open_loop else 1
        self.open_otfb = 0 if open_otfb else 1
        self.open_rffb = 0 if open_rffb else 1
        self.open_tuner = 0 if open_tuner else 1

        self.clamping = clamping

    def generate_white_noise(self, n_points: int):
        r"""Generates white noise"""

        rnd.seed(self.seed1)
        r1 = rnd.random_sample(n_points)
        rnd.seed(self.seed2)
        r2 = rnd.random_sample(n_points)

        return np.exp(2 * np.pi * 1j * r1) * np.sqrt(-2 * np.log(r2))


class SPSOneTurnFeedback(CavityFeedback):
    r"""The SPS one-turn delay feedback and feedforward model in BLonD for a single cavity type.

    Parameters
    ----------
    RFStation : class
        An RFStation type class
    Profile : class
        A Profile type class
    n_sections : int
        Number of sections of the traveling wave cavity
    n_cavities : int
        Number of traveling wave cavities of this type; default is 4
    V_part : float
        Partitioning of the total voltage onto this cavity type; default is 4/9
    G_ff : float
        Feedforward gain; default is 1
    G_llrf : float
        Low-level RF gain; default is 10
    G_tx : float
        Transmitter gain; default is 1
    a_comb : float
        Comb filter coefficient; default is 63/64
    df : float
        Change of the TWC central frequency in Hz from the 2021 measurement; default is 0 Hz
    Commissioning : class
        A SPSCavityLoopCommissioning type class; default is None. If this parameter is None, a new
        SPSCavityLoopCommissioning is used.
    """

    def __init__(
        self,
        RFStation: RFStation,
        Profile: Profile,
        n_sections: int,
        n_cavities: int = 4,
        V_part: float = 4 / 9,
        G_ff: float = 1,
        G_llrf: float = 10,
        G_tx: float = 1,
        a_comb: float = 63 / 64,
        df: float = 0,
        Commissioning: SPSCavityLoopCommissioning = None,
        n_h: int = 0,
    ):
        super().__init__(
            RFStation=RFStation, Profile=Profile, n_cavities=n_cavities, n_s=1, n_h=n_h
        )

        # Set up logging
        self.logger = logging.getLogger(__class__.__name__)

        if Commissioning is None:
            Commissioning = SPSCavityLoopCommissioning()

        # Commissioning options
        self.open_loop = Commissioning.open_loop
        if self.open_loop == 0:  # Open Loop
            self.logger.debug("Opening overall OTFB loop")
        elif self.open_loop == 1:
            self.logger.debug("Closing overall OTFB loop")
        self.open_fb = Commissioning.open_fb
        if self.open_fb == 0:  # Open Feedback
            self.logger.debug("Opening feedback of drive correction")
        elif self.open_fb == 1:
            self.logger.debug("Closing feedback of drive correction")
        self.open_drive = Commissioning.open_drive
        if self.open_drive == 0:  # Open Drive
            self.logger.debug("Opening drive to generator")
        elif self.open_drive == 1:
            self.logger.debug("Closing drive to generator")
        self.open_ff = Commissioning.open_ff
        if self.open_ff == 0:  # Open Feedforward
            self.logger.debug("Opening feed-forward on beam current")
        elif self.open_ff == 1:
            self.logger.debug("Closing feed-forward on beam current")
        self.V_SET = Commissioning.V_SET
        if self.V_SET is None:  # Vset as array or not
            self.set_point_modulation = False
        else:
            self.set_point_modulation = True

        self.cpp_conv = Commissioning.cpp_conv
        self.rot_iq = Commissioning.rot_iq
        self.excitation = Commissioning.excitation

        self.n_sections = int(n_sections)

        self.V_part = float(V_part)
        if self.V_part * (1 - self.V_part) < 0:
            raise RuntimeError(
                "ERROR in SPSOneTurnFeedback: V_part" " should be in range (0,1)!"
            )

        # Gain settings
        self.G_ff = float(G_ff)
        self.G_llrf = float(G_llrf)
        self.G_tx = float(G_tx) / self.n_cavities

        # 200 MHz travelling wave cavity (TWC) model
        if n_sections in [3, 4, 5]:
            self.TWC = eval(
                "SPS" + str(n_sections) + "Section200MHzTWC(" + str(df) + ")"
            )
            if self.open_ff == 1:
                # Feed-forward filter
                self.coeff_ff = getattr(
                    sys.modules[__name__], "feedforward_filter_TWC" + str(n_sections)
                )
                self.n_ff = len(self.coeff_ff)  # Number of coefficients for FF
                self.n_ff_delay = int(
                    0.5 * (self.n_ff - 1)
                    + 0.5
                    * self.TWC.tau
                    / self.rfstation.t_rf[self.n_h, self.rfstation.counter[0]]
                    / 5
                )
                self.logger.debug("Feed-forward delay in samples %d", self.n_ff_delay)

                # Multiply gain by normalisation factors from filter and
                # beam-to generator current
                self.G_ff *= self.TWC.R_beam / (self.TWC.R_gen * np.sum(self.coeff_ff))

        else:
            raise RuntimeError(
                "ERROR in SPSOneTurnFeedback: argument" " n_sections has invalid value!"
            )
        self.logger.debug(
            "SPS OTFB cavities: %d, sections: %d, voltage"
            " partition %.2f, gain: %.2e",
            self.n_cavities,
            n_sections,
            self.V_part,
            self.G_tx,
        )

        # Switch between convolution methods
        if self.cpp_conv:
            self.conv = getattr(self, "call_conv")
        else:
            self.conv = getattr(self, "matr_conv")

        # TWC resonant frequency
        self.omega_c = self.TWC.omega_r
        # Length of arrays in LLRF
        self.n_coarse_ff = int(self.n_coarse / 5)
        # Initialize turn-by-turn variables
        self.dphi_mod = 0

        # Check array length for set point modulation
        if self.set_point_modulation:
            if self.V_SET.shape[0] != 2 * self.n_coarse:
                raise RuntimeError("V_SET length should be %d" % (2 * self.n_coarse))
            self.set_point = getattr(self, "set_point_mod")
        else:
            self.set_point = getattr(self, "set_point_std")
            self.V_SET = np.zeros(2 * self.n_coarse, dtype=complex)

        # Array to hold the bucket-by-bucket voltage with length LLRF
        self.DV_GEN = np.zeros(2 * self.n_coarse, dtype=complex)
        self.logger.debug("Length of arrays on coarse grid 2x %d", self.n_coarse)

        # Array if noise is being injected
        self.NOISE = np.zeros(2 * self.n_coarse, dtype=complex)

        # LLRF MODEL ARRAYS
        # Initialize comb filter
        self.DV_COMB_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.a_comb = float(a_comb)

        # Initialize the delayed signal
        self.DV_DELAYED = np.zeros(2 * self.n_coarse, dtype=complex)

        # Initialize modulated signal (to fr)
        self.DV_MOD_FR = np.zeros(2 * self.n_coarse, dtype=complex)

        # Initialize moving average
        self.n_mov_av = round(
            self.TWC.tau / self.rfstation.t_rf[self.n_h, self.rfstation.counter[0]]
        )
        self.DV_MOV_AVG = np.zeros(2 * self.n_coarse, dtype=complex)
        self.logger.debug("Moving average over %d points", self.n_mov_av)
        if self.n_mov_av < 2:
            raise RuntimeError(
                "ERROR in SPSOneTurnFeedback: profile has to"
                " have at least 12.5 ns resolution!"
            )

        # GENERATOR MODEL ARRAYS
        # Initialize modulated signal (to frf)
        self.DV_MOD_FRF = np.zeros(2 * self.n_coarse, dtype=complex)

        # Initialize induced voltage on coarse grid
        self.V_IND_COARSE_GEN = np.zeros(2 * self.n_coarse, dtype=complex)
        self.CONV_RES = np.zeros(2 * self.n_coarse, dtype=complex)
        self.CONV_PREV = np.zeros(self.n_coarse, dtype=complex)

        # BEAM MODEL ARRAYS
        # Initialize induced beam voltage coarse and fine
        self.V_IND_FINE_BEAM = np.zeros(self.profile.n_slices, dtype=complex)
        self.V_IND_COARSE_BEAM = np.zeros(2 * self.n_coarse, dtype=complex)

        # Initialise feed-forward; sampled every fifth bucket
        if self.open_ff == 1:
            self.logger.debug("Feed-forward active")
            self.I_BEAM_COARSE_FF = np.zeros(2 * self.n_coarse_ff, dtype=complex)
            self.I_BEAM_COARSE_FF_MOD = np.zeros(2 * self.n_coarse_ff, dtype=complex)
            self.I_FF_CORR_MOD = np.zeros(2 * self.n_coarse_ff, dtype=complex)
            self.I_FF_CORR_DEL = np.zeros(2 * self.n_coarse_ff, dtype=complex)
            self.I_FF_CORR = np.zeros(2 * self.n_coarse_ff, dtype=complex)
            self.V_FF_CORR = np.zeros(2 * self.n_coarse_ff, dtype=complex)

        # Update global cavity loop variables before tracking
        self.update_rf_variables()
        self.update_fb_variables()
        self.logger.info("Class initialized")

    def circuit_track(self, no_beam: bool = False):
        r"""Tracking the SPS CL internally."""

        # Update the impulse response at present carrier frequency
        self.TWC.impulse_response_gen(self.omega_carrier, self.rf_centers)
        self.TWC.impulse_response_beam(
            self.omega_carrier, self.profile.bin_centers, self.rf_centers
        )

        # On current measured (I,Q) voltage, apply LLRF model
        self.llrf_model()

        # Generator-induced voltage from generator current
        self.gen_model()

        if not no_beam:
            # Beam-induced voltage from beam profile
            self.beam_model()

        # Sum generator- and beam-induced voltages for coarse grid
        self.V_ANT_START = np.copy(self.V_ANT_COARSE)
        self.V_ANT_COARSE[: self.n_coarse] = self.V_ANT_COARSE[-self.n_coarse :]
        self.V_ANT_COARSE[-self.n_coarse :] = (
            self.V_IND_COARSE_GEN[-self.n_coarse :]
            + self.V_IND_COARSE_BEAM[-self.n_coarse :]
        )

        # Obtain generator-induced voltage on the fine grid by interpolation
        self.V_ANT_FINE_START = np.copy(self.V_ANT_FINE)
        self.V_ANT_FINE[: self.profile.n_slices] = self.V_ANT_FINE[
            -self.profile.n_slices :
        ]
        self.V_ANT_FINE[-self.profile.n_slices :] = self.V_IND_FINE_BEAM[
            -self.profile.n_slices :
        ] + np.interp(
            self.profile.bin_centers,
            self.rf_centers,
            self.V_IND_COARSE_GEN[-self.n_coarse :],
        )

    def llrf_model(self):
        r"""The LLRF model of the SPSOneTurnFeedback. This function calles the functions related
        to the LLRF part of the model in the correct order."""

        # Track all the modules of the LLRF-part of the model
        self.set_point()
        self.error_and_gain()
        self.comb()
        self.one_turn_delay()
        self.mod_to_fr()
        self.mov_avg()

    def gen_model(self):
        r"""The Generator model of the SPSOneTurnFeedback. This function calles the functions related
        to the generator part of the model in the correct order."""

        # Track all the modules for the generator part of the model
        self.mod_to_frf()
        self.sum_and_gain()
        self.gen_response()

    def beam_model(self):
        r"""The Beam model of the SPSOneTurnFeedback. This function find the RF beam current from the Profile-
        object, applies the cavity response towards the beam and the feed-forward correction if engaged.
        """

        # Rotate the RF beam current
        self.I_BEAM_FINE = self.rot_iq * self.I_BEAM_FINE
        self.I_BEAM_COARSE[-self.n_coarse :] = (
            self.rot_iq * self.I_BEAM_COARSE[-self.n_coarse :]
        )

        # Beam-induced voltage
        self.beam_response(coarse=False)
        self.beam_response(coarse=True)

        # Feed-forward
        if self.open_ff == 1:
            # Calculate correction based on previous turn on coarse grid

            # Resample RF beam current to FF sampling frequency
            self.I_BEAM_COARSE_FF[: self.n_coarse_ff] = self.I_BEAM_COARSE_FF[
                -self.n_coarse_ff :
            ]
            I_COARSE_BEAM_RESHAPED = np.copy(self.I_BEAM_COARSE[-self.n_coarse :])
            I_COARSE_BEAM_RESHAPED = I_COARSE_BEAM_RESHAPED.reshape(
                (self.n_coarse_ff, self.n_coarse // self.n_coarse_ff)
            )
            self.I_BEAM_COARSE_FF[-self.n_coarse_ff :] = (
                np.sum(I_COARSE_BEAM_RESHAPED, axis=1) / 5
            )

            self.TWC.impulse_response_ffwd(self.omega_carrier, self.rf_centers[::5])

            # Do a down-modulation to the resonant frequency of the TWC
            self.I_BEAM_COARSE_FF_MOD[: self.n_coarse_ff] = self.I_BEAM_COARSE_FF_MOD[
                -self.n_coarse_ff :
            ]
            self.I_BEAM_COARSE_FF_MOD[-self.n_coarse_ff :] = modulator(
                self.I_BEAM_COARSE_FF[-self.n_coarse_ff :],
                omega_i=self.omega_carrier,
                omega_f=self.omega_c,
                T_sampling=5 * self.T_s,
                phi_0=(self.dphi_mod + self.rfstation.dphi_rf[0]),
            )

            self.I_FF_CORR[: self.n_coarse_ff] = self.I_FF_CORR[-self.n_coarse_ff :]
            self.I_FF_CORR[-self.n_coarse_ff :] = np.zeros(self.n_coarse_ff)
            for ind in range(self.n_coarse_ff, 2 * self.n_coarse_ff):
                for k in range(self.n_ff):
                    self.I_FF_CORR[ind] += (
                        self.coeff_ff[k] * self.I_BEAM_COARSE_FF_MOD[ind - k]
                    )

            # Do a down-modulation to the resonant frequency of the TWC
            phi_delay = (
                self.n_ff_delay * self.T_s * 5 * (self.omega_c - self.omega_carrier)
            )
            self.I_FF_CORR_MOD[: self.n_coarse_ff] = self.I_FF_CORR_MOD[
                -self.n_coarse_ff :
            ]
            self.I_FF_CORR_MOD[-self.n_coarse_ff :] = modulator(
                self.I_FF_CORR[-self.n_coarse_ff :],
                omega_i=self.omega_c,
                omega_f=self.omega_carrier,
                T_sampling=5 * self.T_s,
                phi_0=-(self.dphi_mod + self.rfstation.dphi_rf[0] + phi_delay),
            )

            # Compensate for FIR filter delay
            self.I_FF_CORR_DEL[: self.n_coarse_ff] = self.I_FF_CORR_DEL[
                -self.n_coarse_ff :
            ]
            self.I_FF_CORR_DEL[-self.n_coarse_ff :] = self.I_FF_CORR_MOD[
                self.n_ff_delay : self.n_ff_delay - self.n_coarse_ff
            ]

    # BEAM MODEL
    def beam_response(self, coarse: bool = False):
        r"""Computes the beam-induced voltage on the fine- and coarse-grid by convolving
        the RF beam current with the cavity response towards the beam. The voltage is
        multiplied by the number of cavities to find the total."""
        self.logger.debug("Matrix convolution for V_ind")

        if coarse:
            self.V_IND_COARSE_BEAM[: self.n_coarse] = self.V_IND_COARSE_BEAM[
                -self.n_coarse :
            ]
            self.V_IND_COARSE_BEAM[-self.n_coarse :] = (
                self.n_cavities
                * self.matr_conv(self.I_BEAM_COARSE, self.TWC.h_beam_coarse)[
                    -self.n_coarse :
                ]
                * self.T_s
            )
        else:
            # Only convolve the slices for the current turn because the fine grid points can be less
            # than one turn in length
            self.V_IND_FINE_BEAM[-self.profile.n_slices :] = (
                self.n_cavities
                * self.matr_conv(
                    self.I_BEAM_FINE[-self.profile.n_slices :], self.TWC.h_beam
                )[-self.profile.n_slices :]
                * self.profile.bin_size
            )

    # INDIVIDUAL COMPONENTS ---------------------------------------------------
    # LLRF MODEL

    def set_point_std(self):
        r"""Computes the desired set point voltage in I/Q."""

        self.logger.debug("Entering %s function" % sys._getframe(0).f_code.co_name)
        # Read RF voltage from rf object
        self.V_set = self.set_point_from_rfstation()
        self.V_set = (
            self.n_cavities
            * self.V_part
            * self.V_set
            * np.exp(1j * (-0.5 * np.pi + np.angle(self.rot_iq)))
        )

        # Convert to array
        self.V_SET[: self.n_coarse] = self.V_SET[-self.n_coarse :]
        self.V_SET[-self.n_coarse :] = self.V_set

    def set_point_mod(self):
        r"""This function is called instead of set_point_std if a modulated set point is used.
        That is, if the set point is non-constant over a turn with the periodicity of a turn.
        """

        self.logger.debug("Entering %s function" % sys._getframe(0).f_code.co_name)
        pass

    def error_and_gain(self):
        r"""This function computes the difference between the set point and the antenna voltage
        and amplifies it with the LLRF gain."""

        # Store last turn error signal and update for current turn
        self.DV_GEN[: self.n_coarse] = self.DV_GEN[-self.n_coarse :]
        self.DV_GEN[-self.n_coarse :] = self.G_llrf * (
            self.V_SET[-self.n_coarse :]
            - self.open_loop * self.V_ANT_COARSE[-self.n_coarse :]
            + self.excitation * self.NOISE[-self.n_coarse :]
        )
        self.logger.debug(
            "In %s, average set point voltage %.6f MV",
            sys._getframe(0).f_code.co_name,
            1e-6 * np.mean(np.absolute(self.V_SET)),
        )
        self.logger.debug(
            "In %s, average antenna voltage %.6f MV",
            sys._getframe(0).f_code.co_name,
            1e-6 * np.mean(np.absolute(self.V_ANT_COARSE)),
        )
        self.logger.debug(
            "In %s, average voltage error %.6f MV",
            sys._getframe(0).f_code.co_name,
            1e-6 * np.mean(np.absolute(self.DV_GEN)),
        )

    def comb(self):
        r"""This function applies the comb filter to the error signal."""

        # Shuffle present data to previous data
        self.DV_COMB_OUT[: self.n_coarse] = self.DV_COMB_OUT[-self.n_coarse :]
        # Update present data
        self.DV_COMB_OUT[-self.n_coarse :] = comb_filter(
            self.DV_COMB_OUT[: self.n_coarse],
            self.DV_GEN[-self.n_coarse :],
            self.a_comb,
        )

    def one_turn_delay(self):
        r"""This function applies the complementary delay such that the correction is applied
        with exactly the delay of one turn."""

        # Store last turn delayed signal and compute current turn error signal
        self.DV_DELAYED[: self.n_coarse] = self.DV_DELAYED[-self.n_coarse :]
        self.DV_DELAYED[-self.n_coarse :] = self.DV_COMB_OUT[
            self.n_coarse - self.n_delay : -self.n_delay
        ]

    def mod_to_fr(self):
        r"""This function modulates the error signal to the resonant frequency of the cavity."""

        # Store last turn modulated signal
        self.DV_MOD_FR[: self.n_coarse] = self.DV_MOD_FR[-self.n_coarse :]
        # Note here that dphi_rf is already accumulated somewhere else (i.e. in the tracker).
        self.DV_MOD_FR[-self.n_coarse :] = modulator(
            self.DV_DELAYED[-self.n_coarse :],
            self.omega_carrier,
            self.omega_c,
            self.T_s,
            phi_0=self.dphi_mod,
            dt=self.dT,
        )

    def mov_avg(self):
        r"""This function applies the cavity filter, modelled as a moving average, to the modulated
        error signal."""

        # Store last turn moving average signal
        self.DV_MOV_AVG[: self.n_coarse] = self.DV_MOV_AVG[-self.n_coarse :]
        # Apply moving average filter for current turn
        self.DV_MOV_AVG[-self.n_coarse :] = moving_average(
            self.DV_MOD_FR[-self.n_mov_av - self.n_coarse + 1 :], self.n_mov_av
        )

    # GENERATOR MODEL

    def mod_to_frf(self):
        r"""This function modulates the error signal from the resonant frequency of the cavity to the
        original carrier frequency, the RF frequency."""

        # Store last turn modulated signal
        self.DV_MOD_FRF[: self.n_coarse] = self.DV_MOD_FRF[-self.n_coarse :]
        # Note here that dphi_rf is already accumulated somewhere else (i.e. in the tracker).
        dphi_demod = (self.omega_c - self.omega_carrier) * self.TWC.tau
        self.DV_MOD_FRF[-self.n_coarse :] = self.open_fb * modulator(
            self.DV_MOV_AVG[-self.n_coarse :],
            self.omega_c,
            self.omega_carrier,
            self.T_s,
            phi_0=-(self.dphi_mod + dphi_demod),
            dt=self.dT,
        )

    def sum_and_gain(self):
        r"""Summing of the error signal from the LLRF-part of the model and the set point voltage.
        The generator current is then found by multiplying by the transmitter gain and R_gen. The feed-forward
        current will also be added to the generator current if enabled."""

        # Store generator current signal from the last turn
        self.I_GEN_COARSE[: self.n_coarse] = self.I_GEN_COARSE[-self.n_coarse :]
        # Compute current turn generator current
        self.I_GEN_COARSE[-self.n_coarse :] = (
            self.DV_MOD_FRF[-self.n_coarse :]
            + self.open_drive * self.V_SET[-self.n_coarse :]
        )
        # Apply amplifier gain
        self.I_GEN_COARSE[-self.n_coarse :] *= self.G_tx / self.TWC.R_gen
        if self.open_ff == 1:
            self.I_GEN_COARSE[-self.n_coarse :] = self.I_GEN_COARSE[
                -self.n_coarse :
            ] + self.G_ff * 5 * np.interp(
                self.rf_centers,
                self.rf_centers[::5],
                self.I_FF_CORR_DEL[-self.n_coarse_ff :],
            )

    def gen_response(self):
        r"""Generator current is convolved with cavity response towards the generator to get the
        generator-induced voltage. Multiplied by the number of cavities to find the total generator-
        induced voltage."""

        # Store generator-induced from last turn
        self.V_IND_COARSE_GEN[: self.n_coarse] = self.V_IND_COARSE_GEN[-self.n_coarse :]
        # Compute current turn generator-induced voltage
        self.V_IND_COARSE_GEN[-self.n_coarse :] = (
            self.n_cavities
            * self.matr_conv(self.I_GEN_COARSE, self.TWC.h_gen)[-self.n_coarse :]
            * self.T_s
        )

    def matr_conv(self, I: np.ndarray, h: np.ndarray):
        r"""Convolution of beam current with impulse response; uses a complete
        matrix with off-diagonal elements."""

        return scipy.signal.fftconvolve(I, h, mode="full")[: I.shape[0]]

    def call_conv(self, signal, kernel):
        r"""Routine to call optimised C++ convolution"""

        # Make sure that the buffers are stored contiguously
        signal = np.ascontiguousarray(signal)
        kernel = np.ascontiguousarray(kernel)

        result = np.zeros(len(kernel) + len(signal) - 1)
        bm.convolve(signal, kernel, result=result, mode="full")

        return result

    def update_fb_variables(self):
        r"""Update variables in the feedback"""

        # Phase offset at the end of a 1-turn modulated signal (for demodulated, multiply by -1 as c and r reversed)
        self.phi_mod_0 = (
            (self.omega_carrier_prev - self.omega_c)
            * (self.T_s_prev * self.n_coarse)
            % (2 * np.pi)
        )
        self.dphi_mod += self.phi_mod_0
        self.dphi_mod = self.dphi_mod % (2 * np.pi)

        # Present delay time
        self.n_mov_av = int(self.TWC.tau / self.rfstation.t_rf[self.n_h, self.counter])
        self.n_delay = self.n_coarse - self.n_mov_av

    # Power related functions
    def calc_power(self):
        r"""Method to compute the generator power"""

        return get_power_gen_i(np.copy(self.I_GEN_COARSE), 50)

    def wo_clamping(self):
        pass

    def w_clamping(self):
        pass


class SPSCavityFeedback:
    """Class determining the turn-by-turn total RF voltage and phase correction
    originating from the individual cavity feedbacks. Assumes two 4-section and
    two 5-section travelling wave cavities in the pre-LS2 scenario and four
    3-section and two 4-section cavities in the post-LS2 scenario. The voltage
    partitioning is proportional to the number of sections.

    Parameters
    ----------
    RFStation : class
        An RFStation type class
    Profile : class
        A Profile type class
    G_ff : float or list
        FF gain [1]; if passed as a float, both 3- and 4-section (4- and
        5-section) cavities have the same G_ff in the post- (pre-)LS2
        scenario. If passed as a list, the first and second elements correspond
        to the G_ff of the 3- and 4-section (4- and 5-section) cavity
        feedback in the post- (pre-)LS2 scenario; default is 10
    G_llrf : float or list
        LLRF Gain [1]; convention same as G_ff; default is 10
    G_tx : float or list
        Transmitter gain [1] of the cavity feedback; convention same as G_ff;
        default is 0.5
    a_comb : float
        Comb filter ratio [1]; default is 15/16
    turns :  int
        Number of turns to pre-track without beam
    post_LS2 : bool
        Activates pre-LS2 scenario (False) or post-LS2 scenario (True); default
        is True
    V_part : float
        Voltage partitioning of the shorter cavities; has to be in the range
        (0,1). Default is None and will result in 6/10 for the 3-section
        cavities in the post-LS2 scenario and 4/9 for the 4-section cavities in
        the pre-LS2 scenario
    df : float or list
        Frequency difference between measured frequency and desired frequency;
        same convetion as G_ff; default is 0
    Commissioning : class
        A SPSCavityLoopCommissioning type class; default is None. If this parameter is None, a new
        SPSCavityLoopCommissioning is used.
    """

    def __init__(
        self,
        RFStation: RFStation,
        Profile: Profile,
        G_ff=1,
        G_llrf=10,
        G_tx=0.5,
        a_comb=None,
        turns: int = 1000,
        post_LS2: bool = True,
        V_part=None,
        df=0,
        Commissioning: SPSCavityLoopCommissioning = None,
        n_h: int = 0,
    ):
        # Options for commissioning the feedback
        if Commissioning is None:
            Commissioning = SPSCavityLoopCommissioning()

        self.Commissioning = Commissioning
        self.rot_iq = Commissioning.rot_iq

        self.rfstation = RFStation

        # Parse input for gains
        if hasattr(G_ff, "__iter__"):
            G_ff_1 = G_ff[0]
            G_ff_2 = G_ff[1]
        else:
            G_ff_1 = G_ff
            G_ff_2 = G_ff

        if hasattr(G_llrf, "__iter__"):
            G_llrf_1 = G_llrf[0]
            G_llrf_2 = G_llrf[1]
        else:
            G_llrf_1 = G_llrf
            G_llrf_2 = G_llrf

        if hasattr(G_tx, "__iter__"):
            G_tx_1 = G_tx[0]
            G_tx_2 = G_tx[1]
        else:
            G_tx_1 = G_tx
            G_tx_2 = G_tx

        if hasattr(df, "__iter__"):
            df_1 = df[0]
            df_2 = df[1]
        else:
            df_1 = df
            df_2 = df

        # Voltage partitioning has to be a fraction
        if V_part and V_part * (1 - V_part) < 0:
            raise RuntimeError(
                "SPS cavity feedback: voltage partitioning has to be in the range (0,1)!"
            )

        # Voltage partition proportional to the number of sections
        if post_LS2:
            if not a_comb:
                a_comb = 63 / 64

            if V_part is None:
                V_part = 6 / 10
            self.OTFB_1 = SPSOneTurnFeedback(
                RFStation=RFStation,
                Profile=Profile,
                n_sections=3,
                n_cavities=4,
                V_part=V_part,
                G_ff=float(G_ff_1),
                G_llrf=float(G_llrf_1),
                G_tx=float(G_tx_1),
                a_comb=float(a_comb),
                df=float(df_1),
                Commissioning=self.Commissioning,
                n_h=n_h,
            )
            self.OTFB_2 = SPSOneTurnFeedback(
                RFStation=RFStation,
                Profile=Profile,
                n_sections=4,
                n_cavities=2,
                V_part=1 - V_part,
                G_ff=float(G_ff_2),
                G_llrf=float(G_llrf_2),
                G_tx=float(G_tx_2),
                a_comb=float(a_comb),
                df=float(df_2),
                Commissioning=self.Commissioning,
                n_h=n_h,
            )
        else:
            if not a_comb:
                a_comb = 15 / 16

            if V_part is None:
                V_part = 4 / 9
            self.OTFB_1 = SPSOneTurnFeedback(
                RFStation=RFStation,
                Profile=Profile,
                n_sections=4,
                n_cavities=2,
                V_part=V_part,
                G_ff=float(G_ff_1),
                G_llrf=float(G_llrf_1),
                G_tx=float(G_tx_1),
                a_comb=float(a_comb),
                df=float(df_1),
                Commissioning=self.Commissioning,
                n_h=n_h,
            )
            self.OTFB_2 = SPSOneTurnFeedback(
                RFStation=RFStation,
                Profile=Profile,
                n_sections=5,
                n_cavities=2,
                V_part=1 - V_part,
                G_ff=float(G_ff_2),
                G_llrf=float(G_llrf_2),
                G_tx=float(G_tx_2),
                a_comb=float(a_comb),
                df=float(df_2),
                Commissioning=self.Commissioning,
                n_h=n_h,
            )

        # Set up logging
        self.logger = logging.getLogger(__class__.__name__)

        # Initialise OTFB without beam
        self.turns = int(turns)
        if turns < 1:
            # FeedbackError
            raise RuntimeError(
                "ERROR in SPSCavityFeedback: 'turns' has to" " be a positive integer!"
            )
        self.track_init(debug=Commissioning.debug)

        self.logger.info("Class initialized")

    def track(self):
        r"""Main tracking method for the SPSCavityFeedback. This tracks both cavity types
        with beam."""

        # Track the feedbacks for the two TWC types
        self.OTFB_1.track()
        self.OTFB_2.track()

        # Sum the fine-grid antenna voltage from the TWC types
        self.V_sum = (
            self.OTFB_1.V_ANT_FINE[-self.OTFB_1.profile.n_slices :]
            + self.OTFB_2.V_ANT_FINE[-self.OTFB_2.profile.n_slices :]
        )

        # Convert to amplitude and phase modulation
        self.V_corr, self.alpha_sum = cartesian_to_polar(self.V_sum)

        # Calculate OTFB correction w.r.t. RF voltage and phase in RFStation
        self.V_corr /= self.rfstation.voltage[
            self.OTFB_1.n_h, self.rfstation.counter[0]
        ]
        self.phi_corr = self.alpha_sum - np.angle(
            np.interp(
                self.OTFB_1.profile.bin_centers,
                self.OTFB_1.rf_centers,
                self.OTFB_1.V_SET[-self.OTFB_1.n_coarse :],
            )
        )

    def track_init(self, debug: bool = False):
        r"""Tracking of the SPSCavityFeedback without beam."""

        if debug:
            cmap = plt.get_cmap("jet")
            colors = cmap(np.linspace(0, 1, self.turns))
            plt.figure("Pre-tracking without beam")
            ax = plt.axes([0.18, 0.1, 0.8, 0.8])
            ax.grid()
            ax.set_ylabel("Voltage [V]")

        for i in range(self.turns):
            self.logger.debug("Pre-tracking w/o beam, iteration %d", i)
            self.OTFB_1.track_no_beam()
            if debug:
                ax.plot(
                    self.OTFB_1.profile.bin_centers * 1e6,
                    np.abs(self.OTFB_1.V_ANT_FINE[-self.OTFB_1.profile.n_slices :]),
                    color=colors[i],
                )
                ax.plot(
                    self.OTFB_1.rf_centers * 1e6,
                    np.abs(self.OTFB_1.V_ANT_COARSE[-self.OTFB_1.n_coarse :]),
                    color=colors[i],
                    linestyle="",
                    marker=".",
                )
            self.OTFB_2.track_no_beam()
        if debug:
            plt.show()

        # Interpolate from the coarse mesh to the fine mesh of the beam
        self.V_sum = np.interp(
            self.OTFB_1.profile.bin_centers,
            self.OTFB_1.rf_centers,
            self.OTFB_1.V_IND_COARSE_GEN[-self.OTFB_1.n_coarse :]
            + self.OTFB_2.V_IND_COARSE_GEN[-self.OTFB_2.n_coarse :],
        )

        # Convert to amplitude and phase
        self.V_corr, self.alpha_sum = cartesian_to_polar(self.V_sum)

        # Calculate OTFB correction w.r.t. RF voltage and phase in RFStation
        self.V_corr /= self.rfstation.voltage[
            self.OTFB_1.n_h, self.rfstation.counter[0]
        ]
        self.phi_corr = self.alpha_sum - np.angle(
            np.interp(
                self.OTFB_1.profile.bin_centers,
                self.OTFB_1.rf_centers,
                self.OTFB_1.V_SET[-self.OTFB_1.n_coarse :],
            )
        )


class LHCCavityLoop(CavityFeedback):
    r"""Cavity loop to regulate the RF voltage in the LHC ACS cavities.
    The loop contains a generator, a switch-and-protect device, an RF FB and a
    OTFB. The arrays of the LLRF system cover one turn with exactly one tenth
    of the harmonic (i.e.\ the typical sampling time is about 25 ns).

    Parameters
    ----------
    RFStation : class
        An RFStation type class
    Profile : class
        Beam profile object
    n_cavities : int
        Number of cavities per beam; default is 8
    f_c : float
        Central cavity frequency [Hz]; default is 400.789e6 Hz
    G_gen : float
        Overall driver chain gain [1]; default is 1
    I_gen_offset : float
        Generator current offset [A]; default is 0
    n_pretrack : int
        Number of turns to pre-track without beam; default is 200
    Q_L : float
        Cavity loaded quality factor; default is 20000
    R_over_Q : float
        Cavity R/Q [Ohm]; default is 45 Ohms
    tau_loop : float
        Total loop delay [s]; default is 650e-9 s
    tau_otfb : float
        Total loop delay as seen by OTFB [s]; default is 1472e-9 s
    RFFB : class
        LHCCavityLoopCommissioning type class containing RF FB gains and delays. If this parameter is None, a new
        LHCCavityLoopCommissioning is used.
    """

    def __init__(
        self,
        RFStation: RFStation,
        Profile: Profile,
        n_cavities: int = 8,
        f_c: float = 400.789e6,
        G_gen: float = 1,
        I_gen_offset: float = 0,
        n_pretrack: int = 200,
        Q_L: float = 20000,
        R_over_Q: float = 45,
        tau_loop: float = 650e-9,
        tau_otfb: float = 1472e-9,
        RFFB: LHCCavityLoopCommissioning = None,
        n_h: int = 0,
    ):
        super().__init__(
            RFStation=RFStation, Profile=Profile, n_cavities=n_cavities, n_s=10, n_h=n_h
        )

        # Set up logging
        self.logger = logging.getLogger(__class__.__name__)

        # Options for commissioning the feedback
        if RFFB is None:
            RFFB = LHCCavityLoopCommissioning()

        # Import classes and parameters
        self.RFFB = RFFB
        self.I_gen_offset = I_gen_offset
        self.G_gen = G_gen
        self.n_pretrack = n_pretrack
        self.omega_c = 2 * np.pi * f_c
        # TODO: implement optimum loaded Q
        self.Q_L = Q_L
        self.R_over_Q = R_over_Q
        self.tau_loop = tau_loop
        self.tau_otfb = tau_otfb
        self.logger.debug("Cavity loaded Q is %.0f", self.Q_L)

        # Import RF FB properties
        self.open_drive = self.RFFB.open_drive
        self.open_drive_inv = self.RFFB.open_drive_inv
        self.open_loop = self.RFFB.open_loop
        self.open_otfb = self.RFFB.open_otfb
        self.open_rffb = self.RFFB.open_rffb
        self.open_tuner = self.RFFB.open_tuner
        self.clamping = self.RFFB.clamping
        self.alpha = self.RFFB.alpha
        self.d_phi_ad = self.RFFB.d_phi_ad
        self.G_a = self.RFFB.G_a
        self.G_d = self.RFFB.G_d
        self.G_o = self.RFFB.G_o
        self.tau_a = self.RFFB.tau_a
        self.tau_d = self.RFFB.tau_d
        self.tau_o = self.RFFB.tau_o
        self.mu = self.RFFB.mu
        self.power_thres = self.RFFB.power_thres
        self.v_swap_thres = (
            np.sqrt(2 * self.power_thres / (self.R_over_Q * self.Q_L)) / self.G_gen
        )
        self.excitation = self.RFFB.excitation
        self.excitation_otfb_1 = self.RFFB.excitation_otfb_1
        self.excitation_otfb_2 = self.RFFB.excitation_otfb_2

        self.logger.debug("Length of arrays in generator path %d", self.n_coarse)

        # Initialise FIR filter for OTFB
        self.fir_n_taps = 63
        self.fir_coeff = fir_filter_lhc_otfb_coeff(n_taps=self.fir_n_taps)
        self.logger.debug("Sum of FIR coefficients %.4e" % np.sum(self.fir_coeff))

        self.update_rf_variables()
        self.update_fb_variables()
        self.logger.debug("Relative detuning is %.4e", self.detuning)

        # Arrays
        self.V_EXC = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_FB_IN = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_AC_IN = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_AN_IN = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_AN_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_DI_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_OTFB = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_OTFB_INT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_FIR_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_FB_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_SWAP_OUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.I_TEST = np.zeros(2 * self.n_coarse, dtype=complex)
        self.TUNER_INPUT = np.zeros(2 * self.n_coarse, dtype=complex)
        self.TUNER_INTEGRATED = np.zeros(2 * self.n_coarse, dtype=complex)

        self.V_ANT_FINE = np.zeros(self.profile.n_slices + 1, dtype=complex)
        self.I_GEN_FINE = np.zeros(self.profile.n_slices + 1, dtype=complex)

        # Pre-track without beam
        self.logger.debug("Track without beam for %d turns", self.n_pretrack)
        if self.excitation:
            self.excitation_otfb = False
            self.logger.debug("Injecting noise in voltage set point")
            self.track_no_beam_excitation(self.n_pretrack)
        elif self.excitation_otfb_1 or self.excitation_otfb_2:
            self.excitation_otfb = True
            self.logger.debug("Injecting noise at OTFB output")
            self.track_no_beam_excitation_otfb(self.n_pretrack)
        else:
            self.excitation_otfb = False
            self.logger.debug("Pre-tracking without beam")
            self.track_no_beam(self.n_pretrack)

        self.logger.info("LHCCavityLoop class initialized")

    def circuit_track(self, no_beam: bool = False):
        r"""Track the feedback model"""
        if not no_beam:
            self.I_BEAM_FINE *= -1j * np.exp(
                1j * (self.rfstation.phi_s[self.rfstation.counter[0]])
            )
            self.I_BEAM_COARSE[-self.n_coarse :] *= -1j * np.exp(
                1j * (self.rfstation.phi_s[self.rfstation.counter[0]])
            )

        # Track the different parts of the model
        self.update_arrays()
        self.update_set_point()
        self.track_one_turn()

        if not no_beam:
            # Resample generator current to the fine-grid
            self.I_GEN_FINE = np.interp(
                np.concatenate(
                    (
                        np.array([self.profile.bin_centers[0] - self.profile.bin_size]),
                        self.profile.bin_centers,
                    )
                ),
                self.rf_centers,
                self.I_GEN_COARSE[-self.n_coarse :],
            )

            # Compute the fine-grid antenna voltage through solving a sparse matrix equation
            self.cavity_response_fine_matrix()

            # Apply the tuner correction
            self.tuner()

    def cavity_response(self, samples):
        r"""ACS cavity reponse model"""

        self.V_ANT_COARSE[self.ind] = (
            self.I_GEN_COARSE[self.ind - 1] * self.R_over_Q * samples
            + self.V_ANT_COARSE[self.ind - 1]
            * (1 - 0.5 * samples / self.Q_L + 1j * self.detuning * samples)
            - self.I_BEAM_COARSE[self.ind - 1] * 0.5 * self.R_over_Q * samples
        )

    def cavity_response_fine_matrix(self):
        r"""ACS cavity response model in matrix form on the fine-grid"""

        # Number of samples on fine grid
        self.samples_fine = self.omega_rf * self.profile.bin_size

        # Find initial value of antenna voltage and generator current
        t_at_init = self.profile.bin_centers[0] - self.profile.bin_size
        V_A_init = interp1d(
            np.concatenate(
                (self.rf_centers - self.T_s * self.n_coarse, self.rf_centers)
            ),
            self.V_ANT_COARSE,
            fill_value="extrapolate",
        )(t_at_init)
        I_gen_init = interp1d(
            np.concatenate(
                (self.rf_centers - self.T_s * self.n_coarse, self.rf_centers)
            ),
            self.I_BEAM_COARSE,
            fill_value="extrapolate",
        )(t_at_init)

        self.V_ANT_FINE = cavity_response_sparse_matrix(
            I_beam=self.I_BEAM_FINE,
            I_gen=self.I_GEN_FINE,
            n_samples=self.profile.n_slices,
            V_ant_init=V_A_init,
            I_gen_init=I_gen_init,
            samples_per_rf=self.samples_fine,
            R_over_Q=self.R_over_Q,
            Q_L=self.Q_L,
            detuning=self.detuning,
        )

        self.V_ANT_FINE[-self.profile.n_slices :] = (
            self.n_cavities * self.V_ANT_FINE[-self.profile.n_slices :]
        )

    def generator_current(self):
        r"""Generator response

        Attributes
        I_TEST : complex array
            Test point for open loop measurements (when injecting a generator
            offset)
        """

        # From V_swap_out in closed loop, constant in open loop
        # TODO: missing terms for changing voltage and beam current
        self.I_TEST[self.ind] = self.G_gen * self.V_SWAP_OUT[self.ind]
        self.I_GEN_COARSE[self.ind] = (
            self.open_drive * self.I_TEST[self.ind]
            + self.open_drive_inv * self.I_gen_offset
        )

    def generator_power(self):
        r"""Calculation of generator power from generator current"""

        return 0.5 * self.R_over_Q * self.Q_L * np.absolute(self.I_GEN_COARSE) ** 2

    def one_turn_feedback(self, T_s: float):
        r"""Apply effect of the OTFB on the analog branch"""

        # OTFB itself
        self.V_OTFB_INT[self.ind] = (
            self.alpha * self.V_OTFB_INT[self.ind - self.n_coarse]
            + self.G_o
            * (1 - self.alpha)
            * self.V_AC_IN[self.ind - self.n_coarse + self.n_otfb]
        )

        # FIR filter
        self.V_FIR_OUT[self.ind] = self.fir_coeff[0] * self.V_OTFB_INT[self.ind]
        for k in range(1, self.fir_n_taps):
            self.V_FIR_OUT[self.ind] += (
                self.fir_coeff[k] * self.V_OTFB_INT[self.ind - k]
            )

        # AC coupling at output
        self.V_OTFB[self.ind] = (
            (1 - T_s / self.tau_o) * self.V_OTFB[self.ind - 1]
            + self.V_FIR_OUT[self.ind]
            - self.V_FIR_OUT[self.ind - 1]
        )

    def rf_feedback(self, T_s: float):
        r"""Analog and digital RF feedback response"""

        # Calculate voltage difference to act on
        self.V_FB_IN[self.ind] = (
            self.V_SET[self.ind - self.n_delay]
            - self.open_loop * self.V_ANT_COARSE[self.ind - self.n_delay]
        )

        # On the analog branch, OTFB can contribute
        self.V_AC_IN[self.ind] = (
            (1 - T_s / self.tau_o) * self.V_AC_IN[self.ind - 1]
            + self.V_FB_IN[self.ind]
            - self.V_FB_IN[self.ind - 1]
        )
        self.one_turn_feedback(T_s=T_s)

        self.V_AN_IN[self.ind] = (
            self.V_FB_IN[self.ind]
            + self.open_otfb * self.V_OTFB[self.ind]
            + int(bool(self.excitation_otfb)) * self.V_EXC[self.ind]
        )

        # Output of analog feedback (separate branch)
        self.V_AN_OUT[self.ind] = self.V_AN_OUT[self.ind - 1] * (
            1 - T_s / self.tau_a
        ) + self.G_a * (self.V_AN_IN[self.ind] - self.V_AN_IN[self.ind - 1])

        # Output of digital feedback (separate branch)
        self.V_DI_OUT[self.ind] = (
            self.V_DI_OUT[self.ind - 1] * (1 - T_s / self.tau_d)
            + T_s
            / self.tau_d
            * self.G_a
            * self.G_d
            * np.exp(1j * self.d_phi_ad)
            * self.V_FB_IN[self.ind - 1]
        )

        # Total output: sum of analog and digital feedback
        self.V_FB_OUT[self.ind] = self.open_rffb * (
            self.V_AN_OUT[self.ind] + self.V_DI_OUT[self.ind]
        )

    def update_set_point(self):
        r"""Updates the set point for the next turn based on the design RF
        voltage."""
        coeff = np.polyfit(
            [0, self.n_coarse + 1],
            [self.V_SET[-self.n_coarse], self.set_point_from_rfstation()[0]],
            1,
        )
        poly = np.poly1d(coeff)
        v_set_prev = poly(np.linspace(0, self.n_coarse, self.n_coarse))

        self.V_SET = np.concatenate((v_set_prev, self.set_point_from_rfstation()))

    def swap(self):
        r"""Model of the Switch and Protect module: clamping of the output
        power above a given input power."""

        # TODO: check implementation
        if self.clamping:
            self.V_SWAP_OUT[self.ind] = (
                self.v_swap_thres
                * smooth_step(
                    np.abs(self.V_FB_OUT[self.ind]), x_max=self.v_swap_thres, N=0
                )
                * np.exp(1j * np.angle(self.V_FB_OUT[self.ind]))
            )
        else:
            self.V_SWAP_OUT[self.ind] = self.V_FB_OUT[self.ind]

    def tuner(self):
        r"""Model of the tuner algorithm."""

        # Compute the detuning factor for the current turn
        dtune = (
            -(self.mu / 2)
            * (
                np.min(self.TUNER_INTEGRATED[-self.n_coarse :].imag)
                + np.max(self.TUNER_INTEGRATED[-self.n_coarse :].imag)
            )
            / (self.rfstation.voltage[self.n_h, self.counter] / self.n_cavities) ** 2
        )

        # Propagate the corrections to the detuning two the global parameters
        self.detuning = self.detuning + dtune * self.open_tuner
        self.d_omega = self.detuning * self.omega_c
        self.omega_c = self.omega_rf + self.d_omega

    def tuner_input(self):
        r"""Gathering data for the detuning algortithm"""

        # Calculating input signal
        self.TUNER_INPUT[self.ind] = self.I_GEN_COARSE[self.ind] * np.conj(
            self.V_ANT_COARSE[self.ind]
        )

        # Apply CIC-component
        self.TUNER_INTEGRATED[self.ind] = (
            (1 / 64)
            * (
                self.TUNER_INPUT[self.ind]
                - 2 * self.TUNER_INPUT[self.ind - 8]
                + self.TUNER_INPUT[self.ind - 16]
            )
            + 2 * self.TUNER_INTEGRATED[self.ind - 1]
            - self.TUNER_INTEGRATED[self.ind - 2]
        )

    def track_one_turn(self):
        r"""Single-turn tracking, index by index."""

        for i in range(self.n_coarse):
            T_s = self.T_s
            self.ind = i + self.n_coarse
            self.cavity_response(samples=T_s * self.omega_rf)
            self.rf_feedback(T_s=T_s)
            self.swap()
            self.generator_current()
            self.tuner_input()

    def update_arrays(self):
        r"""Moves the array indices by one turn (n_coarse points) from the
        present turn to prepare the next turn. All arrays except for V_SET."""

        self.V_ANT_COARSE = np.concatenate(
            (self.V_ANT_COARSE[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_FB_IN = np.concatenate(
            (self.V_FB_IN[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_AC_IN = np.concatenate(
            (self.V_AC_IN[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_AN_IN = np.concatenate(
            (self.V_AN_IN[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_AN_OUT = np.concatenate(
            (self.V_AN_OUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_DI_OUT = np.concatenate(
            (self.V_DI_OUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_OTFB = np.concatenate(
            (self.V_OTFB[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_OTFB_INT = np.concatenate(
            (self.V_OTFB_INT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_FIR_OUT = np.concatenate(
            (self.V_FIR_OUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_FB_OUT = np.concatenate(
            (self.V_FB_OUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.V_SWAP_OUT = np.concatenate(
            (self.V_SWAP_OUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.I_GEN_COARSE = np.concatenate(
            (self.I_GEN_COARSE[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.I_TEST = np.concatenate(
            (self.I_TEST[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.TUNER_INPUT = np.concatenate(
            (self.TUNER_INPUT[self.n_coarse :], np.zeros(self.n_coarse, dtype=complex))
        )
        self.TUNER_INTEGRATED = np.concatenate(
            (
                self.TUNER_INTEGRATED[self.n_coarse :],
                np.zeros(self.n_coarse, dtype=complex),
            )
        )

    def update_fb_variables(self):
        r"""Update counter and frequency-dependent variables in a given turn"""

        # Delay time
        self.n_delay = round(self.tau_loop / self.T_s)
        self.n_fir = round(0.5 * (self.fir_n_taps - 1))
        self.n_otfb = round(self.tau_otfb / self.T_s) + self.n_fir

        # Present detuning
        self.d_omega = self.omega_c - self.omega_rf

        # Dimensionless quantities
        self.samples = self.omega_rf * self.T_s
        self.detuning = self.d_omega / self.omega_c

    def update_set_point_excitation(self, excitation: np.ndarray, turn: int):
        r"""Updates the set point for the next turn based on the excitation to
        be injected."""

        self.V_SET = np.concatenate(
            (
                self.V_SET[self.n_coarse :],
                excitation[turn * self.n_coarse : (turn + 1) * self.n_coarse],
            )
        )

    def track_no_beam_excitation(self, n_turns: int):
        r"""Pre-tracking for n_turns turns, without beam. With excitation; set
        point from white noise. V_EXC_IN and V_EXC_OUT can be used to measure
        the transfer function of the system at set point.

        Attributes
        ----------
        V_EXC_IN : complex array
            Noise being played in set point; n_coarse * n_turns elements
        V_EXC_OUT : complex array
            System reaction to noise (accumulated from V_ANT); n_coarse * n_turns
            elements
        """

        self.V_EXC_IN = 1000 * self.RFFB.generate_white_noise(self.n_coarse * n_turns)
        self.V_EXC_OUT = np.zeros(self.n_coarse * n_turns, dtype=complex)
        self.V_SET = np.concatenate(
            (np.zeros(self.n_coarse, dtype=complex), self.V_EXC_IN[0 : self.n_coarse])
        )
        self.track_one_turn()
        self.V_EXC_OUT[0 : self.n_coarse] = self.V_ANT_COARSE[
            self.n_coarse : 2 * self.n_coarse
        ]
        for n in range(1, n_turns):
            self.update_arrays()
            self.update_set_point_excitation(self.V_EXC_IN, n)
            self.track_one_turn()
            self.V_EXC_OUT[
                n * self.n_coarse : (n + 1) * self.n_coarse
            ] = self.V_ANT_COARSE[self.n_coarse : 2 * self.n_coarse]

    def track_no_beam_excitation_otfb(self, n_turns: int):
        r"""Pre-tracking for n_turns turns, without beam. With excitation; set
        point from white noise. V_EXC_IN and V_EXC_OUT can be used to measure
        the transfer function of the system at otfb.

        Attributes
        ----------
        V_EXC_IN : complex array
            Noise being played in set point; n_coarse * n_turns elements
        V_EXC_OUT : complex array
            System reaction to noise (accumulated from V_ANT); n_coarse * n_turns
            elements
        """

        self.V_EXC_IN = 10000 * self.RFFB.generate_white_noise(self.n_coarse * n_turns)
        self.V_EXC_OUT = np.zeros(self.n_coarse * n_turns, dtype=complex)
        self.V_SET = np.zeros(2 * self.n_coarse, dtype=complex)
        self.V_EXC = np.concatenate(
            (np.zeros(self.n_coarse, dtype=complex), self.V_EXC_IN[0 : self.n_coarse])
        )

        self.track_one_turn()
        if self.excitation_otfb_1:
            self.V_EXC_OUT[: self.n_coarse] = self.V_FB_IN[
                self.n_coarse : 2 * self.n_coarse
            ]
        elif self.excitation_otfb_2:
            self.V_EXC_OUT[: self.n_coarse] = self.V_OTFB[self.ind]
        for n in range(1, n_turns):
            self.update_arrays()
            self.V_EXC = np.concatenate(
                (
                    np.zeros(self.n_coarse, dtype=complex),
                    self.V_EXC_IN[n * self.n_coarse : (n + 1) * self.n_coarse],
                )
            )

            for i in range(self.n_coarse):
                self.ind = i + self.n_coarse
                self.cavity_response(self.T_s * self.omega_rf)
                self.rf_feedback(self.T_s)
                self.swap()
                self.generator_current()
                if self.excitation_otfb_1:
                    self.V_EXC_OUT[n * self.n_coarse + i] = self.V_FB_IN[
                        self.n_coarse + i
                    ]
                elif self.excitation_otfb_2:
                    self.V_EXC_OUT[n * self.n_coarse + i] = self.V_OTFB[self.ind]

    @staticmethod
    def half_detuning(imag_peak_beam_current, R_over_Q, rf_frequency, voltage):
        """Optimum detuning for half-detuning scheme

        Parameters
        ----------
        imag_peak_beam_current : float
            Peak RF beam current
        R_over_Q : float
            Cavity R/Q
        rf_frequency : float
            RF frequency
        voltage : float
            RF voltage amplitude in the cavity

        Returns
        -------
        float
            Optimum detuning (revolution) frequency in the half-detuning scheme
        """

        return -0.25 * R_over_Q * imag_peak_beam_current / voltage * rf_frequency

    @staticmethod
    def half_detuning_power(peak_beam_current, voltage):
        """RF power consumption half-detuning scheme with optimum detuning

        Parameters
        ----------
        peak_beam_current : float
            Peak RF beam current
        voltage : float
            Cavity voltage

        Returns
        -------
        float
            Optimum detuning (revolution) frequency in the half-detuning scheme
        """

        return 0.125 * peak_beam_current * voltage

    @staticmethod
    def optimum_Q_L(detuning, rf_frequency):
        """Optimum loaded Q when no real part of RF beam current is present

        Parameters
        ----------
        detuning : float
            Detuning frequency
        rf_frequency : float
            RF frequency

        Returns
        -------
        float
            Optimum loaded Q
        """

        return np.fabs(0.5 * rf_frequency / detuning)

    @staticmethod
    def optimum_Q_L_beam(R_over_Q, real_peak_beam_current, voltage):
        """Optimum loaded Q when a real part of RF beam current is present

        Parameters
        ----------
        peak_beam_current : float
            Peak RF beam current
        R_over_Q : float
            Cavity R/Q
        voltage : float
            Cavity voltage

        Returns
        -------
        float
            Optimum loaded Q
        """

        return voltage / (R_over_Q * real_peak_beam_current)



'''
BLonD math and physics core functions

@author Stefan Hegglin, Konstantinos Iliakis, Panagiotis Tsapatsaris, Georgios Typaldos
@date 20.10.2017
'''

import numpy as np

from ..utils import butils_wrap_cpp as _cpp
from ..utils import butils_wrap_python as _py
from . import precision


def use_cpp():
    '''
    Replace all python functions by there equivalent in cpp
    '''
    # dictionary storing the CPP versions of the most compute intensive functions #
    cpp_func_dict = {
        'rfft': np.fft.rfft,
        'irfft': np.fft.irfft,
        'rfftfreq': np.fft.rfftfreq,

        'kick': _cpp.kick,
        'rf_volt_comp': _cpp.rf_volt_comp,
        'drift': _cpp.drift,
        'slice_beam': _cpp.slice_beam,
        'slice_smooth': _cpp.slice_smooth,
        'linear_interp_kick': _cpp.linear_interp_kick,
        'synchrotron_radiation': _cpp.synchrotron_radiation,
        'synchrotron_radiation_full': _cpp.synchrotron_radiation_full,
        'music_track': _cpp.music_track,
        'music_track_multiturn': _cpp.music_track_multiturn,
        'fast_resonator': _cpp.fast_resonator,
        'beam_phase': _cpp.beam_phase,
        'beam_phase_fast': _cpp.beam_phase_fast,
        'sparse_histogram': _cpp.sparse_histogram,
        'distribution_from_tomoscope': _cpp.distribution_from_tomoscope,
        'set_random_seed': _cpp.set_random_seed,

        'sin_cpp': _cpp.sin_cpp,
        'cos_cpp': _cpp.cos_cpp,
        'exp_cpp': _cpp.exp_cpp,
        'mean_cpp': _cpp.mean_cpp,
        'std_cpp': _cpp.std_cpp,
        'where_cpp': _cpp.where_cpp,
        'interp_cpp': _cpp.interp_cpp,
        'interp_const_space': _cpp.interp_const_space,
        'interp_const_bin': _cpp.interp_const_bin,
        'cumtrapz': _cpp.cumtrapz,
        'trapz_cpp': _cpp.trapz_cpp,
        'linspace_cpp': _cpp.linspace_cpp,
        'argmin_cpp': _cpp.argmin_cpp,
        'argmax_cpp': _cpp.argmax_cpp,
        'convolve': _cpp.convolve,
        'arange_cpp': _cpp.arange_cpp,
        'sum_cpp': _cpp.sum_cpp,
        'sort_cpp': _cpp.sort_cpp,
        'add_cpp': _cpp.add_cpp,
        'mul_cpp': _cpp.mul_cpp,
        'random_normal': _cpp.random_normal,

        'device': 'CPU_CPP'
    }

    # add numpy functions in the dictionary
    for fname in dir(np):
        if callable(getattr(np, fname)) and (fname not in cpp_func_dict) \
                and (fname[0] != '_'):
            cpp_func_dict[fname] = getattr(np, fname)

    # add basic numpy modules to dictionary as they are not callable
    cpp_func_dict['random'] = getattr(np, 'random')
    cpp_func_dict['fft'] = getattr(np, 'fft')

    __update_active_dict(cpp_func_dict)
    # print('---------- Using the C++ computational backend ----------')


def use_numba():
    '''
    Replace all python functions by their equivalent in numba
    '''

    from blond.utils import butils_wrap_numba as _nu


    # dictionary storing the Numba-only versions of the most compute intensive functions #
    nu_func_dict = {
        'rfft': np.fft.rfft,
        'irfft': np.fft.irfft,
        'rfftfreq': np.fft.rfftfreq,

        'kick': _nu.kick,
        'rf_volt_comp': _nu.rf_volt_comp,
        'drift': _nu.drift,
        'slice_beam': _nu.slice_beam,
        'slice_smooth': _nu.slice_smooth,
        'linear_interp_kick': _nu.linear_interp_kick,
        'synchrotron_radiation': _nu.synchrotron_radiation,
        'synchrotron_radiation_full': _nu.synchrotron_radiation_full,
        'music_track': _nu.music_track,
        'music_track_multiturn': _nu.music_track_multiturn,
        'fast_resonator': _nu.fast_resonator,
        'beam_phase': _nu.beam_phase,
        'beam_phase_fast': _nu.beam_phase_fast,
        'sparse_histogram': _nu.sparse_histogram,
        'distribution_from_tomoscope': _nu.distribution_from_tomoscope,
        'set_random_seed': _nu.set_random_seed,

        'device': 'CPU_NU'
    }

    # add numpy functions in the dictionary
    for fname in dir(np):
        if callable(getattr(np, fname)) and (fname not in nu_func_dict) \
                and (fname[0] != '_'):

            nu_func_dict[fname] = getattr(np, fname)

    # add basic numpy modules to dictionary as they are not callable
    nu_func_dict['random'] = getattr(np, 'random')
    nu_func_dict['fft'] = getattr(np, 'fft')

    # Update the global functions
    __update_active_dict(nu_func_dict)

    # print('---------- Using the Numba computational backend ----------')


def use_py():
    '''
    Replace all python functions by there equivalent in python
    '''

    # dictionary storing the Python-only versions of the most compute intensive functions #
    py_func_dict = {
        'rfft': np.fft.rfft,
        'irfft': np.fft.irfft,
        'rfftfreq': np.fft.rfftfreq,

        'kick': _py.kick,
        'rf_volt_comp': _py.rf_volt_comp,
        'drift': _py.drift,
        'slice_beam': _py.slice_beam,
        'slice_smooth': _py.slice_smooth,
        'linear_interp_kick': _py.linear_interp_kick,
        'synchrotron_radiation': _py.synchrotron_radiation,
        'synchrotron_radiation_full': _py.synchrotron_radiation_full,
        'music_track': _py.music_track,
        'music_track_multiturn': _py.music_track_multiturn,
        'fast_resonator': _py.fast_resonator,
        'beam_phase': _py.beam_phase,
        'beam_phase_fast': _py.beam_phase_fast,
        'sparse_histogram': _py.sparse_histogram,
        'distribution_from_tomoscope': _py.distribution_from_tomoscope,
        'set_random_seed': _py.set_random_seed,

        'device': 'CPU_PY'
    }

    # add numpy functions in the dictionary
    for fname in dir(np):
        if callable(getattr(np, fname)) and (fname not in py_func_dict) \
                and (fname[0] != '_'):

            py_func_dict[fname] = getattr(np, fname)

    # add basic numpy modules to dictionary as they are not callable
    py_func_dict['random'] = getattr(np, 'random')
    py_func_dict['fft'] = getattr(np, 'fft')

    # Update the global functions
    __update_active_dict(py_func_dict)

    # print('---------- Using the Python computational backend ----------')



def use_cpu():
    '''
    If not library is found, use the python implementations
    '''
    # from .. import get_libblond
    if _cpp.get_libblond() is None:
        try: # try to use numba
            use_numba()
        except ImportError as e:
            print(f"Error: {e}. Please install numba with 'pip install numba' to use the numba backend."
                  f" Using the python backend instead.")
            use_py()
    else:
        use_cpp()


def use_mpi():
    '''
    Replace some bm functions with MPI implementations
    '''

    mpi_func_dict = {
        'device': 'CPU_MPI'
    }
    globals().update(mpi_func_dict)


def in_mpi():
    """Check if we are currently in MPI mode

    Returns:
        bool: True if in MPI mode
    """
    return globals()['device'] == 'CPU_MPI'


def use_fftw():
    '''
    Replace the existing rfft and irfft implementations
    with the ones coming from _cpp.
    '''
    print('---------- Using the FFTW FFT library ----------')

    fftw_func_dict = {
        'rfft': _cpp.rfft,
        'irfft': _cpp.irfft,
        'rfftfreq': _cpp.rfftfreq
    }
    globals().update(fftw_func_dict)


# precision can be single or double
def use_precision(_precision='double'):
    """Change the precision used in caclulations.

    Args:
        _precision (str, optional): Can be either 'single' or 'double'. Defaults to 'double'.
    """
    print(f'---------- Using {_precision} precision numeric datatypes ----------')
    precision.set(_precision)

    try:
        from ..gpu import GPU_DEV
        GPU_DEV.load_library(_precision)
    except Exception as e:
        # The GPU backend is not available
        pass
    _cpp.load_libblond(_precision)


def __update_active_dict(new_dict):
    '''
    Update the currently active dictionary. Removes the keys of the currently
    active dictionary from globals() and spills the keys
    from new_dict to globals()
    Args:
        new_dict A dictionary which contents will be spilled to globals()
    '''
    if not hasattr(__update_active_dict, 'active_dict'):
        __update_active_dict.active_dict = new_dict

    # delete all old implementations/references from globals()
    for key in __update_active_dict.active_dict.keys():
        if key in globals():
            del globals()[key]
    # for key in globals().keys():
    #     if key in __update_active_dict.active_dict.keys():
    #         del globals()[key]
    # add the new active dict to the globals()
    globals().update(new_dict)
    __update_active_dict.active_dict = new_dict


# GPU Related Utilities
def get_gpu_device():
    """Get the GPU device object

    Returns:
        _type_: _description_
    """
    from ..gpu import GPU_DEV
    return GPU_DEV


def use_gpu(gpu_id=0):
    """Use the GPU device to perform the calculations.

    Args:
        gpu_id (int, optional): The device id. Defaults to 0.
    """
    if gpu_id < 0:
        return

    from ..gpu import GPU_DEV

    GPU_DEV.set(gpu_id)

    import cupy as cp

    from ..gpu import butils_wrap_cupy as _cupy

    gpu_func_dict = {
        'rfft': cp.fft.rfft,
        'irfft': cp.fft.irfft,
        'rfftfreq': cp.fft.rfftfreq,
        'convolve': cp.convolve,
        # 'convolve': _cupy.convolve,
        'beam_phase': _cupy.beam_phase,
        'beam_phase_fast': _cupy.beam_phase_fast,
        'kick': _cupy.kick,
        'rf_volt_comp': _cupy.rf_volt_comp,
        'drift': _cupy.drift,
        'linear_interp_kick': _cupy.linear_interp_kick,
        # 'LIKick_n_drift': _cupy.linear_interp_kick_drift,
        'synchrotron_radiation': _cupy.synchrotron_radiation,
        'synchrotron_radiation_full': _cupy.synchrotron_radiation_full,
        'slice_beam': _cupy.slice_beam,
        # 'interp_const_space': _cupy.interp,
        'interp_const_space': cp.interp,
        'device': 'GPU'
    }
    # add cupy functions in the dictionary
    for fname in dir(cp):
        if callable(getattr(cp, fname)) and (fname not in gpu_func_dict):
            gpu_func_dict[fname] = getattr(cp, fname)
    __update_active_dict(gpu_func_dict)

    # add basic cupy modules to dictionary as they are not callable
    gpu_func_dict['random'] = getattr(cp, 'random')
    gpu_func_dict['fft'] = getattr(cp, 'fft')

    # print('---------- Using the GPU computational backend ----------')
    # print(f'---------- GPU Device: id {GPU_DEV.id}, name {GPU_DEV.name}, Compute Capability {GPU_DEV.dev.compute_capability} ----------',
    #        flush=True)


def report_backend():
    """Prints the currently active backend
    """
    global device
    print(f'---------- Using the {device} computational backend ----------')

###############################################################################
# By default use the CPU backend (python-only or C++)
use_cpu()
###############################################################################




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to plot different bunch features**

:Authors: **Helga Timko**, **Danilo Quartullo**
'''

from __future__ import division

from builtins import range

import matplotlib.pyplot as plt
import numpy as np

from ..impedances.impedance_sources import InputTable
import matplotlib.lines as mlines

# Automatic layouting of plots
plt.rc('figure', autolayout=True)
plt.rc('savefig', bbox='tight')
plt.rc('savefig', pad_inches=0.1)
plt.rc('figure', figsize = [8,6])

def plot_impedance_vs_frequency(induced_voltage_freq, figure_index = 0,
                                plot_total_impedance = True, plot_spectrum=False,
                                plot_interpolated_impedances = False, style='-',
                                cut_left_right=None, cut_up_down=None,
                                dirname='fig', show_plots = False):
    """
    Plots Impedance in frequency domain. Given an InducedVoltageFreq object, i.e. one
    that was created from a list of impedances:
     - can plot either the total, summed impedance or each individual one from the list
     - can additionally plot the beam spectrum
     - and can plot the impedances on an interpolated frequency axis, if the table
        given to induced_voltage_time has been interpolated on a new axis

    Additionally, either outputs the plots in the console, or save them on disk.

    :param induced_voltage_freq:  InducedVoltageFreq object
        InducedVoltageFreq object which impedances are to be plotted
    :param figure_index:  int
        Index of the figure, e.g. the turn number if creating a plot for every turn
        Default: 0
    :param plot_total_impedance: bool
        If True, plots the total, summed impedance which is calculated internally in the
        induced_voltage_time.
        If False, plots the raw impedances from the table given to induced_voltage_time
        Default: True
    :param plot_spectrum:  bool
        If True, plots the beam spectrum in addition to the impedances.
        If False, no beam spectrum is plotted
        Default: False
    :param plot_interpolated_impedances: bool
        If True, the impedance table in induced_voltage_time must have been interpolated
        to a new frequency axis. Will then plot the impedances on that new frequency axis.
        If False, plots the raw impedances on the original frequency axis
        given by the table in induced_voltage_time
        Default: False
    :param style: str
        Matplotlib plot style string
        Default: "-"
    :param cut_left_right: 2-tuple of floats
        x-axis limits of the plots (frequency in Hz)
        Default: None
    :param cut_up_down: 2-tuple of floats
        y-axis limits of the plots (resistance in Ohm)
        Default: None
    :param dirname: str
        Path to directory in which the plots are to be saved.
        Default: "fig"
    :param show_plots: bool
        If True, will output the plots directly in the console
        If False, will save the plots to directory given in dirname
    """

    if plot_total_impedance:

        ax1 = plt.subplots()[1]
        # Renormalize Impedances to the cont. Fourier Transform values, i.e. Ohm.
        # In the InducedVoltage Objects they are normalized to the DFT values by deviding
        # by bin size. This is reverted here when plotting to get sensible units

        ax1.plot(induced_voltage_freq.freq,
                 induced_voltage_freq.total_impedance.real * induced_voltage_freq.profile.bin_size, style, label = "Real Impedance")
        ax1.plot(induced_voltage_freq.freq,
                 induced_voltage_freq.total_impedance.imag * induced_voltage_freq.profile.bin_size, style, label = "Imaginary Impedance")

        ax1.set_xlim(cut_left_right)
        ax1.set_ylim(cut_up_down)
        ax1.set_xlabel("Frequency [Hz]")
        ax1.set_ylabel("Impedance [Ohm]")
        if plot_spectrum:
            ax2 = ax1.twinx()
            ax2.plot(induced_voltage_freq.profile.beam_spectrum_freq,
                     np.abs(induced_voltage_freq.profile.beam_spectrum), label = "Beam Spectrum")

            ax2.set_xlim(cut_left_right)
            ax2.set_ylim(cut_up_down)
            ax2.set_ylabel("Beam Spectrum [a.u.]")
        if plot_spectrum:
            ax2.legend()
        else:
            ax1.legend()
        fign = dirname + '/sum_imp_vs_freq_fft' "%d" % figure_index + '.png'

        if show_plots:
            plt.show()
        else:
            plt.savefig(fign)
            plt.clf()

    else:

        fig0 = plt.figure(0)
        ax0 = fig0.add_subplot(111)

        fig1 = plt.figure(1)
        ax1 = fig1.add_subplot(111)
        # The impedance sources themselves are properly normalized already.
        for i in range(len(induced_voltage_freq.impedance_source_list)):
            if isinstance(induced_voltage_freq.impedance_source_list[i],
                          InputTable) and not plot_interpolated_impedances:
                ax0.plot(induced_voltage_freq.impedance_source_list[i].frequency_array_loaded,
                         induced_voltage_freq.impedance_source_list[i].Re_Z_array_loaded, style)
                ax0.set_xlim(cut_left_right)
                ax0.set_ylim(cut_up_down)
                ax1.plot(induced_voltage_freq.impedance_source_list[i].frequency_array_loaded,
                         induced_voltage_freq.impedance_source_list[i].Im_Z_array_loaded, style)
                ax1.set_xlim(cut_left_right)
                ax1.set_ylim(cut_up_down)
            elif plot_interpolated_impedances:
                ax0.plot(induced_voltage_freq.impedance_source_list[i].frequency_array,
                         induced_voltage_freq.impedance_source_list[i].impedance.real, style)
                ax0.set_xlim(cut_left_right)
                ax0.set_ylim(cut_up_down)
                ax1.plot(induced_voltage_freq.impedance_source_list[i].frequency_array,
                         induced_voltage_freq.impedance_source_list[i].impedance.imag, style)
                ax1.set_xlim(cut_left_right)
                ax1.set_ylim(cut_up_down)

        if plot_interpolated_impedances:
            fig_suffix = "interpolated_freq"
        else:
            fig_suffix = "table_freq"

        ax0.set_xlabel("Frequency [Hz]")
        ax0.set_ylabel("Real impedance [Ohm]")

        ax1.set_xlabel("Frequency [Hz]")
        ax1.set_ylabel("Imaginary Impedance [Ohm]")


        fign1 = dirname + '/real_imp_vs_' + fig_suffix + '_' "%d" % figure_index + '.png'
        if plot_spectrum:
            ax2 = ax0.twinx()
            spectrum, = ax2.plot(induced_voltage_freq.profile.beam_spectrum_freq,
                     np.abs(induced_voltage_freq.profile.beam_spectrum), label = "Beam Spectrum",color = "r")
            ax2.set_xlim(cut_left_right)
            ax2.set_ylim(cut_up_down)
            ax2.set_ylabel("Beam Spectrum [a.u.]")
            ax2.legend(handles=[spectrum])



        plt.figure(0)
        if show_plots:
            plt.show()
        else:
            plt.savefig(fign1)
            plt.clf()
        fign2 = dirname + '/imag_imp_vs_' + fig_suffix + '_' "%d" % figure_index + '.png'


        if plot_spectrum:
            ax3 = ax1.twinx()
            spectrum, = ax3.plot(induced_voltage_freq.profile.beam_spectrum_freq,
                     np.abs(induced_voltage_freq.profile.beam_spectrum), label = "Beam Spectrum", color = "r")
            ax3.set_xlim(cut_left_right)
            ax3.set_ylim(cut_up_down)
            ax3.set_ylabel("Beam Spectrum [a.u.]")

            ax3.legend(handles=[spectrum])
        plt.figure(1)
        if show_plots:
            plt.show()
        else:
            plt.savefig(fign2)
            plt.clf()

def plot_wake_vs_time(induced_voltage_time, figure_index = 0,
                      plot_total_wake = True,
                      plot_interpolated_wake = False, style='-',
                      cut_left_right=None, cut_up_down=None,
                      dirname='fig', show_plots = False):
    """
    Plots wakes in time domain. Given an InducedVoltageTime object, i.e. one
    that was created from a list of wakes:
     - can plot either the total, summed wake or each individual one from the list
     - and can plot the wakes  on an interpolated wake axis, if the table
        given to induced_voltage_time has been interpolated on a new axis

    Additionally, either outputs the plots in the console, or save them on disk.

    :param induced_voltage_time:  InducedVoltageTime object
        InducedVoltageTime object which wakes are to be plotted
    :param figure_index:  int
        Index of the figure, e.g. the turn number if creating a plot for every turn
        Default: 0
    :param plot_total_wake: bool
        If True, plots the total, summed wake which is calculated internally in the
        induced_voltage_time.
        If False, plots the raw wakes from the table of wakes given to induced_voltage_time
        Default: True
    :param plot_interpolated_wake: bool
        If True, the wake table in induced_voltage_time must have been interpolated
        to a new frequency axis. Will then plot the wakes on that new frequency axis.
        If False, plots the raw wakes on the original frequency axis
        given by the table in induced_voltage_time
        Default: False
    :param style: str
        Matplotlib plot style string
        Default: "-"
    :param cut_left_right: 2-tuple of floats
        x-axis limits of the plots (frequency in Hz)
        Default: None
    :param cut_up_down: 2-tuple of floats
        y-axis limits of the plots (resistance in Ohm)
        Default: None
    :param dirname: str
        Path to directory in which the plots are to be saved.
        Default: "fig"
    :param show_plots: bool
        If True, will output the plots directly in the console
        If False, will save the plots to directory given in dirname
    """

    if plot_total_wake:

        ax1 = plt.subplots()[1]

        ax1.plot(induced_voltage_time.time,
                 induced_voltage_time.total_wake, style, label ="Wake")

        ax1.set_xlim(cut_left_right)
        ax1.set_ylim(cut_up_down)
        ax1.set_xlabel("Time [s]")
        ax1.set_ylabel("Wake [Ohm/s]")
        ax1.legend()
        fign = dirname + '/sum_wake_vs_table_times' "%d" % figure_index + '.png'

        if show_plots:
            plt.show()
        else:
            plt.savefig(fign)
            plt.clf()

    else:

        fig0 = plt.figure(0)
        ax0 = fig0.add_subplot(111)

        for i in range(len(induced_voltage_time.wake_source_list)):
            if isinstance(induced_voltage_time.wake_source_list[i],
                          InputTable) and not plot_interpolated_wake:
                ax0.plot(induced_voltage_time.wake_source_list[i].time_array,
                         induced_voltage_time.wake_source_list[i].wake_array, style)
                ax0.set_xlim(cut_left_right)
                ax0.set_ylim(cut_up_down)
            elif plot_interpolated_wake:
                ax0.plot(induced_voltage_time.wake_source_list[i].new_time_array,
                         induced_voltage_time.wake_source_list[i].wake, style)
                ax0.set_xlim(cut_left_right)
                ax0.set_ylim(cut_up_down)

        if plot_interpolated_wake:
            fig_suffix = "interpolated_times"
        else:
            fig_suffix = "table_times"

        ax0.set_xlabel("Time [s]")
        ax0.set_ylabel("Wake [Ohm/s]")


        fign1 = dirname + '/wake_vs_' + fig_suffix + '_' "%d" % figure_index + '.png'



        plt.figure(0)
        if show_plots:
            plt.show()
        else:
            plt.savefig(fign1)
            plt.clf()

def plot_induced_voltage_vs_bin_centers(total_induced_voltage, style='-', figure_index = 0,
                                        dirname='fig', show_plots = False):
    """
    Plots the total induced voltage calculated in the TotalInducedVoltage object given.

    :param total_induced_voltage: TotalInducedVoltage object
        TotalInducedVoltage object containing the total induced voltage to be plotted
    :param style: str
        Matplotlib style string
        Default: "-"
    :param figure_index: int
        Index of the figure, e.g. the turn number if creating a plot for every turn
        Default: 0
    :param dirname: str
        Directory to save the plots to
        Default: "fig"
    :param show_plots: bool
        If True, outputs the plots directly to console
        If False, saved the plots to disk in given directory
        Default: False
    """

    fig0 = plt.figure(0)
    fig0.set_size_inches(8, 6)
    ax0 = plt.axes()
    plt.plot(total_induced_voltage.profile.bin_centers, total_induced_voltage.induced_voltage, style)
    ax0.set_xlabel("Time [s]")
    ax0.set_ylabel("Induced voltage [V]")

    # Save plot
    fign = dirname + '/induced_voltage_' "%d" % figure_index + '.png'
    if show_plots:
        plt.show()
    else:
        plt.savefig(fign)
        plt.clf()



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Sanity check. Run before committing**

:Authors: **Helga Timko, Konstantinos Iliakis**
'''

import argparse
import subprocess
import sys
import textwrap


def get_modified_files():
    """Returns a list containing all the modified or commited files tracked by git.

    Returns:
        list: list of file names
    """
    output = b''
    ret = subprocess.run(['git diff --no-commit-id --name-only -r HEAD'],
                         check=True, capture_output=True, shell=True)
    output += ret.stdout
    ret = subprocess.run(['git diff-tree --no-commit-id --name-only -r HEAD'],
                         check=True, capture_output=True, shell=True)
    output += ret.stdout
    files = output.decode().splitlines()
    files = [f for f in files if f.endswith('.py')]
    return files


class SanityCheck:
    """SanityCheck class
    """
    print_prompt = '==== SANITY_CHECK:'
    isort_settings = '--check --diff'
    flake8_settings = '--ignore E501,W503,W504,W291'
    pylint_settings = '-f colorized'
    pytest_settings = '-v'
    coverage_settings = '--cov'

    def __init__(self, all_tools=False, docs=False, pytest=None,
                 flake8=None, pylint=None, isort=None, coverage=False):
        """Constructor.

        Args:
            all_tools (bool, optional): _description_. Defaults to False.
            docs (bool, optional): _description_. Defaults to False.
            pytest (_type_, optional): _description_. Defaults to None.
            flake8 (_type_, optional): _description_. Defaults to None.
            pylint (_type_, optional): _description_. Defaults to None.
            isort (_type_, optional): _description_. Defaults to None.
            coverage (bool, optional): _description_. Defaults to False.
        """
        if all_tools:
            docs = True
            if flake8 is None:
                flake8 = './blond'
            if pytest is None:
                pytest = './unittests'
            if pylint is None:
                pylint = './blond'
            if isort is None:
                isort = './blond'

        if docs:
            self.compile_docs()
        if flake8:
            self.run_flake8(flake8)
        if pylint:
            self.run_pylint(pylint)
        if isort:
            self.run_isort(isort)
        if pytest:
            self.run_pytest(pytest, coverage)

        print(f"{self.print_prompt} Finished all checks!")

    def compile_docs(self):
        """Compile the documentation in html format.
        """
        print(f"{self.print_prompt} Compiling the documentation")

        ret = subprocess.run(['make -C __doc html'], shell=True, check=False,
                             capture_output=False)

        if ret.returncode != 0:
            print(f"{self.print_prompt} Documentation compilation failed")
        else:
            print(f"{self.print_prompt} Documentation compiled\n")

    def run_pylint(self, pylint):
        """Use pylint to report potential code syntax issues.

        Args:
            pylint (_type_): _description_
        """
        print(f"{self.print_prompt} Using pylint to report code syntax issues")

        files = ''

        if pylint == 'git':
            print(f"{self.print_prompt} Executing pylint on committed/modified files")
            files = ' '.join(get_modified_files())
        elif isinstance(pylint, str):
            files = pylint

        subprocess.run([f'{sys.executable} -m pylint {self.pylint_settings} {files}'],
                       shell=True, check=False, capture_output=False)

    def run_flake8(self, flake8):
        """Use flake8 to report code styling issues.

        Args:
            flake8 (_type_): _description_
        """
        print(f"{self.print_prompt} Using flake8 to report code style issues")

        files = ''

        if flake8 == 'git':
            print(f"{self.print_prompt} Executing flake8 on committed/modified files")
            files = ' '.join(get_modified_files())
        elif isinstance(flake8, str):
            files = flake8

        subprocess.run([f'{sys.executable} -m flake8 {self.flake8_settings} {files}'],
                       shell=True, check=False, capture_output=False)

    def run_isort(self, isort):
        """Use isort to sort imports.

        Args:
            isort (bool): _description_
        """
        print(f"{self.print_prompt} Using isort to sort the imports")

        files = ''

        if isort == 'git':
            print(f"{self.print_prompt} Executing isort on committed/modified files")
            files = ' '.join(get_modified_files())
        elif isinstance(isort, str):
            files = isort

        subprocess.run([f'{sys.executable} -m isort {self.isort_settings} {files}'],
                       shell=True, check=False, capture_output=False)

    def run_pytest(self, pytest, coverage):
        """Use pytest to run the unittests.

        Args:
            pytest (str): A space separated list of unittest names to run.
            coverage (bool): If True, report code coverage.
        """
        print(
            f"{self.print_prompt} Running the unittests to test the code's correctness")

        files = pytest

        if coverage:
            settings = f'{self.coverage_settings} {self.pytest_settings}'
        else:
            settings = self.pytest_settings

        subprocess.run([f'{sys.executable} -m pytest {settings} {files}'],
                       shell=True, check=False, capture_output=False)


def main():
    """Main function
    """
    # Arguments read from command line
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''
        SANITY CHECKER; run before committing from BLonD folder
        E.g. > python sanity_check.py -flake8 "llrf/signal_processing.py beam/profile.py"
             > python sanity_check.py -pytest ../unittests/general
             > python sanity_check.py --all
             > python sanity_check.py -pylint git (pylint report only committed or modified files)
             > python sanity_check.py -isort (use isort to sort the imports of the entirey BLonD library)
        '''))

    parser.add_argument('-all', '--all', dest='all', action='store_true',
                        help='Execute all checks', default=False)

    parser.add_argument('-docs', '--docs', dest='docs', action='store_true',
                        help='Compile docs in html format', default=False)

    parser.add_argument('-pytest', '--pytest', dest='pytest', const='./unittests',
                        nargs='?', type=str, default=None,
                        help='Run all unit-tests (default) or only ' +
                        'unit-tests found in the given files/directories')

    parser.add_argument('-coverage', '--coverage', dest='coverage',
                        action='store_true', default=False,
                        help='Report code coverage. Requires specifying the -pytest option.')

    parser.add_argument('-flake8', '--flake8', dest='flake8', const='./blond',
                        nargs='?', type=str, default=None,
                        help='Run the flake8 tool; on the entire BLonD library (default),' +
                        ' or on the specified files' +
                        ' or on committed/ modified files with -flake8 git')

    parser.add_argument('-pylint', '--pylint', dest='pylint', const='./blond',
                        nargs='?', type=str, default=None,
                        help='Run the pylint tool; on the entire BLonD library (default),' +
                        ' or on the specified files' +
                        ' or on committed/ modified files with -pylint git')

    parser.add_argument('-isort', '--isort', dest='isort', const='./blond',
                        nargs='?', type=str, default=None,
                        help='Run the isort tool; on the entire BLonD library (default),' +
                        ' or on the specified files' +
                        ' or on committed/ modified files with -isort git')

    args = parser.parse_args()

    # Passing a default option if no argument is given
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    # Call the actual sanity check
    SanityCheck(args.all, args.docs, args.pytest, args.flake8,
                args.pylint, args.isort, args.coverage)


if __name__ == "__main__":
    main()



# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
:Authors: **Danilo Quartullo**, **Joel Repond**
'''

from __future__ import division, print_function

import numpy as np
import scipy.special as scisp
from scipy.constants import e


def analytical_gaussian_resonator(sigma_t, Q, R_s, omega_r, tau_array, n_particles):
    r"""Calculate the analytical induced voltage for a gaussian bunch and a resonator.

    Parameters
    ----------
    sigma_t : float
        RMS of the beam longitudinal coordinates [s].
    Q : float
        Quality factor of the resonator [1].
    R_s : float
        Shunt impedance of the resonator [:math:`\Omega`].
    omega_r : float
        Angular resonant frequency [rad/s].
    tau_array : float array
        Time array [s] for which the induced voltage should be calculated.
    n_particles : int
        Beam intensity [1].

    Returns
    -------
    induced_voltage : float array
        Induced voltage [V] (multiplied by -1 according to BLonD conventions).

    Notes
    -----
    The formula can be derived using the following lines of codes in
    Mathematica:

    .. code-block:: mathematica

        lambda = Exp[-(tau - t)^2 / (2 * sigma^2)] / ((2*Pi)^0.5 * sigma)
        wake = 2 * alpha * Rs * Exp[-alpha * t] * (Cos[ombar * t] - alpha / ombar
                                                   * Sin[ombar * t])
        output = Integrate[wake * lambda, {t, 0, Infinity},
                           Assumptions -> {sigma > 0, alpha > 0, Rs > 0, ombar > 0, tau in Reals}]
        outputreal = Simplify[ComplexExpand[Re[output]],
                              {sigma > 0, alpha > 0, Rs > 0, ombar > 0, tau in  Reals}]


    The imaginary part of output is identically equal to zero even if
    Mathematica cannot see that. To verify that the expression is correct
    launch:

    .. code-block:: mathematica

        output2 = Integrate[wake * lambda, t]
        outputreal2 = Simplify[ComplexExpand[Re[output2]],
                               {sigma > 0, alpha > 0, Rs > 0, ombar > 0, tau in Reals, t > 0}]
        Simplify[ComplexExpand[Re[d/dt output2]],
                 {sigma > 0, alpha > 0, Rs > 0, ombar > 0, tau in Reals, t > 0}]


    and check that applying the fundamental calculus theorem to outputreal2,
    one obtains back outputreal; the formula [5] in

    H. E. Salzer, "Formulas for Calculating the Error Function of a Complex
    Variable", Mathematical Tables and Other Aids to Computation, Vol. 5,
    No. 34 (Apr., 1951),pp. 67-70

    can be useful.

    """
    alpha = omega_r / (2 * Q)
    ombar = np.sqrt(omega_r**2 - alpha**2)

    A = (alpha * sigma_t**2 - tau_array + 1j * ombar * sigma_t**2) / (np.sqrt(2) * sigma_t)
    B = alpha * ombar * sigma_t**2 - ombar * tau_array
    result = R_s * alpha / ombar * np.e**(0.5 * (alpha**2 - ombar**2) * sigma_t**2 - alpha * tau_array) *\
        (scisp.erfc(A).real * (ombar * np.cos(B) + alpha * np.sin(B)) +
         scisp.erfc(A).imag * (alpha * np.cos(B) - ombar * np.sin(B)))

    induced_voltage = -n_particles * e * result

    return induced_voltage



# coding: utf8
# Copyright 2014-2017 CERN. This software is distributed under the
# terms of the GNU General Public License version 3 (GPL Version 3),
# copied verbatim in the file LICENSE.md.
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Function(s) for pre-processing input data**

:Authors: **Helga Timko**, **Alexandre Lasheen**, **Danilo Quartullo**,
    **Simon Albright**
'''

from __future__ import division

from builtins import range, str

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import splev, splrep

from ..plots.plot import fig_folder


class RFStationOptions:
    r""" Class to preprocess the RF data (voltage, phase, harmonic) for
    RFStation, interpolating it to every turn.

    Parameters
    ----------
    interpolation : str
        Interpolation options for the data points. Available options are
        'linear' (default) and 'cubic'
    smoothing : float
        Smoothing value for 'cubic' interpolation
    plot : bool
        Option to plot interpolated arrays; default is False
    figdir : str
        Directory to save optional plot; default is 'fig'
    figname : list of str
        Figure name to save optional plot; default is 'data', different arrays
        will have figures with different indices
    sampling : int
        Decimation value for plotting; default is 1

    """

    def __init__(self, interpolation='linear', smoothing=0, plot=False,
                 figdir='fig', figname=['data'], sampling=1):

        if interpolation in ['linear', 'cubic']:
            self.interpolation = str(interpolation)
        else:
            # InterpolationError
            raise RuntimeError(
                "ERROR: Interpolation scheme in" +
                " RFStationOptions not recognised. Aborting...")

        self.smoothing = float(smoothing)

        if isinstance(plot, bool):
            self.plot = bool(plot)
        else:
            # TypeError
            raise RuntimeError("ERROR: plot value in PreprocessRamp" +
                               " not recognised. Aborting...")

        self.figdir = str(figdir)
        self.figname = figname  # str(figname)

        if sampling > 0:
            self.sampling = int(sampling)
        else:
            # TypeError
            raise RuntimeError("ERROR: sampling value in PreprocessRamp" +
                               " not recognised. Aborting...")

    def reshape_data(self, input_data, n_turns, n_rf, interp_time,
                     t_start=0):
        r"""Checks whether the user input is consistent with the expectation
        for the RFStation object. The possibilites are detailed in the
        documentation of the RFStation object.

        Parameters
        ----------
        input_data : Ring.synchronous_data, Ring.alpha_0,1,2
            Main input data to reshape
        n_turns : RFStation.n_turns
            Number of turns the simulation should be. Note that if
            the input_data is passed as a tuple it is expected that the
            input_data is a program. This parameter is only relevant if the
            rf program is passed as an array/list with the rigth size
            (see interp_time otherwise)
        n_rf : RFStation.n_rf
            The number of rf harmonics in the station. The simulation is
            stopped if the input_data shape does not correspond to the expected
            number of rf harmonics.
        interp_time : Ring.cycle_time
            Ensure that the rf program is interpolated on the same time basis
            as the Ring.momentum program
        t_start : Ring.RingOptions.t_start or float
            Uses the same t_start as the one used to interpolate the momentum
            program in Ring. This value can nevertheless be changed to a custom
            value if necessary.

        Returns
        -------
        output_data
            Returns the data with the adequate shape for the RStation object

        """

        # TO BE IMPLEMENTED: if you pass a filename the function reads the file
        # and reshape the data
        if isinstance(input_data, str):
            pass

        # If single float, expands the value to match the input number of turns
        # and rf harmonics
        if isinstance(input_data, float) or isinstance(input_data, int):
            output_data = input_data * np.ones((n_rf, n_turns + 1))

        # If tuple, separate time and synchronous data and check data
        elif isinstance(input_data, tuple):

            output_data = []

            # Hot fix to safely treat t_start
            if t_start is None:
                t_start = 0

            interp_time = interp_time + t_start

            # If there is only one rf harmonic, it is expected that the user
            # passes a tuple with (time, data). However, the user can also pass
            # a tuple which size is the number of section as ((time, data), ).
            # and this if condition takes this into account
            if (n_rf == 1) and (len(input_data) > 1):
                input_data = (input_data, )

            if len(input_data) != n_rf:
                # InputDataError
                raise RuntimeError("ERROR in RFStation: the input data " +
                                   "does not match the number of rf harmonics")

            # Loops over all the rf harmonics to interpolate the programs,
            # appends the results on the output_data list which is afterwards
            # converted to a numpy.array
            for index_rf in range(n_rf):
                input_data_time = input_data[index_rf][0]
                input_data_values = input_data[index_rf][1]

                if len(input_data_values) \
                        != len(input_data_time):
                    # InputDataError
                    raise RuntimeError("ERROR in RFStation: synchronous " +
                                       "data does not match the time data")

                if self.interpolation == 'linear':
                    output_data.append(np.interp(interp_time,
                                                 input_data_time,
                                                 input_data_values))
                elif self.interpolation == 'cubic':
                    interp_funtion = splrep(input_data_time, input_data_values,
                                            s=self.smoothing)
                    output_data.append(splev(interp_time, interp_funtion))

            output_data = np.array(output_data, ndmin=2, dtype=float)

            # Plot original and interpolated data
            if self.plot:
                # Directory where plots will be stored
                fig_folder(self.figdir)

                # Plot
                for index_rf in range(n_rf):
                    input_data_time = input_data[index_rf][0]
                    input_data_values = input_data[index_rf][1]

                    plt.figure('RFStationOptions', figsize=(8, 6))
                    plt.clf()
                    ax = plt.axes([0.15, 0.1, 0.8, 0.8])
                    ax.plot(interp_time[::self.sampling],
                            output_data[index_rf][::self.sampling],
                            label='Interpolated data')
                    ax.plot(input_data_time, input_data_values, '.',
                            label='Input data', color='r')
                    ax.set_xlabel('Time [s]')
                    ax.set_ylabel("%s" % self.figname[index_rf])
                    ax.legend = plt.legend(
                        bbox_to_anchor=(0., 1.02, 1., .102),
                        loc=3, ncol=2, mode='expand', borderaxespad=0.)

                    # Save figure
                    fign = self.figdir + '/preprocess_' "%s" % self.figname[index_rf] + \
                        '.png'
                    plt.savefig(fign)

        # If array/list, compares with the input number of turns and
        # if synchronous_data is a single value converts it into a (n_turns+1)
        # array
        elif isinstance(input_data, np.ndarray) or \
                isinstance(input_data, list):

            input_data = np.array(input_data, ndmin=2, dtype=float)
            output_data = np.zeros((n_rf, n_turns + 1), dtype=float)

            # If the number of points is exactly the same as n_rf, this means
            # that the rf program for each harmonic is constant, reshaping
            # the array so that the size is [n_sections,1] for successful
            # reshaping
            if input_data.size == n_rf:
                input_data = input_data.reshape((n_rf, 1))

            if len(input_data) != n_rf:
                # InputDataError
                raise RuntimeError("ERROR in RFStation: the input data " +
                                   "does not match the number of rf harmonics")

            for index_rf in range(len(input_data)):
                if len(input_data[index_rf]) == 1:
                    output_data[index_rf] = input_data[index_rf] * \
                        np.ones(n_turns + 1)

                elif len(input_data[index_rf]) == (n_turns + 1):
                    output_data[index_rf] = np.array(
                        input_data[index_rf])

                else:
                    # InputDataError
                    raise RuntimeError("ERROR in Ring: The input data " +
                                       "does not match the proper length " +
                                       "(n_turns+1)")

        return output_data


def combine_rf_functions(function_list, merge_type='linear', resolution=1e-3,
                         Ring=None, main_h=True):
    r"""Function to combine different RF programs. Each program is passed in a
    tuple with complete function (single valued or numpy array) and 2-list
    [start_time, stop_time].

    Parameters
    ----------
    function_list : list of tuples
        each tuple has form (function, [start_time, stop_time])
        function can be a numpy.ndarray of format [time, value] or single valued
        if function is single valued it will be assumed constant from start_time to stop_time
        if function is numpy.ndarray it will be truncated to start_time, stop_time
    merge_type : str
        string signifying type of merge available, options are:
            linear : function will be linearly interpolated from function_1[stop_time] to function_2[start_time]
            isoadiabatic : designed for voltage functions and intended to maintain adiabaticity during change of voltage, best suited to flat momentum sections
            linear_tune : for use with voltages, provides a linear change in the tune from function_1[stop_time] to function_2[start_time]
    resolution : float
        the time in seconds between points of the interpolation
    Ring : class
        A Ring type class, only used with linear_tune merge_type
    main_h : boolean
        if main_h is True dE is considered in linear_tune merge_type, otherwise dE is set to 0

    Returns
    -------
    2 dimensional numpy.ndarray containing [time, value] of merged functions

    """

    nFunctions = len(function_list)

    if not isinstance(merge_type, list):
        merge_type = (nFunctions - 1) * [merge_type]
    if not isinstance(resolution, list):
        resolution = (nFunctions - 1) * [resolution]

    if len(merge_type) != nFunctions:
        # InputDataError
        raise RuntimeError("ERROR: merge_type list wrong length")
    if len(resolution) != nFunctions:
        # InputDataError
        raise RuntimeError("ERROR: resolution list wrong length")

    timePoints = []
    for i in range(nFunctions):
        timePoints += function_list[i][1]
    if not np.all(np.diff(timePoints)) > 0:
        # InputDataError
        raise RuntimeError("ERROR: in combine_rf_functions, times are not" +
                           " monotonically increasing!")

    fullFunction = []
    fullTime = []

    # Determines if 1st function is single valued or array and stores values
    if not isinstance(function_list[0][0], np.ndarray):
        fullFunction += 2 * [function_list[0][0]]
        fullTime += function_list[0][1]

    else:
        start = np.where(function_list[0][0][0] > function_list[0][1][0])[0][0]
        stop = np.where(function_list[0][0][0] > function_list[0][1][1])[0][0]

        funcTime = [function_list[0][1][0]] + \
            function_list[0][0][0][start:stop].tolist() + \
            [function_list[0][1][1]]
        funcProg = np.interp(funcTime, function_list[0][0][0],
                             function_list[0][0][1])

        fullFunction += funcProg.tolist()
        fullTime += funcTime

    # Loops through remaining functions merging them as requested and
    # storing results
    for i in range(1, nFunctions):

        if merge_type[i - 1] == 'linear':

            if not isinstance(function_list[i][0], np.ndarray):
                fullFunction += 2 * [function_list[i][0]]
                fullTime += function_list[i][1]

            else:
                start = np.where(function_list[i][0][0] >=
                                 function_list[i][1][0])[0][0]
                stop = np.where(function_list[i][0][0] >=
                                function_list[i][1][1])[0][0]

                funcTime = [function_list[i][1][0]] + \
                    function_list[i][0][0][start:stop].tolist() + \
                    [function_list[i][1][1]]
                funcProg = np.interp(funcTime, function_list[i][0][0],
                                     function_list[i][0][1])

                fullFunction += funcProg.tolist()
                fullTime += funcTime

        elif merge_type[i - 1] == 'isoadiabatic':

            if not isinstance(function_list[i][0], np.ndarray):

                tDur = function_list[i][1][0] - fullTime[-1]
                Vinit = fullFunction[-1]
                Vfin = function_list[i][0]
                k = (1. / tDur) * (1 - (1. * Vinit / Vfin)**0.5)

                nSteps = int(tDur / resolution[i - 1])
                time = np.linspace(float(fullTime[-1]),
                                   float(function_list[i][1][0]), nSteps)
                volts = Vinit / ((1 - k * (time - time[0]))**2)

                fullFunction += volts.tolist() + 2 * [function_list[i][0]]
                fullTime += time.tolist() + function_list[i][1]

            else:

                start = np.where(function_list[i][0][0] >=
                                 function_list[i][1][0])[0][0]
                stop = np.where(function_list[i][0][0] >=
                                function_list[i][1][1])[0][0]

                funcTime = [function_list[i][1][0]] + \
                    function_list[i][0][0][start:stop].tolist() + \
                    [function_list[i][1][1]]
                funcProg = np.interp(funcTime, function_list[i][0][0],
                                     function_list[i][0][1])

                tDur = funcTime[0] - fullTime[-1]
                Vinit = fullFunction[-1]
                Vfin = funcProg[0]
                k = (1. / tDur) * (1 - (1. * Vinit / Vfin)**0.5)

                nSteps = int(tDur / resolution[i - 1])
                time = np.linspace(float(fullTime[-1]), float(funcTime[0]), nSteps)
                volts = Vinit / ((1 - k * (time - time[0]))**2)

                fullFunction += volts.tolist() + funcProg.tolist()
                fullTime += time.tolist() + funcTime

        elif merge_type[i - 1] == 'linear_tune':

            # harmonic, charge and 2pi are constant so can be ignored
            if not isinstance(function_list[i][0], np.ndarray):

                initPars = Ring.parameters_at_time(fullTime[-1])
                finalPars = Ring.parameters_at_time(function_list[i][1][0])

                vInit = fullFunction[-1]
                vFin = function_list[i][0]

                if main_h is False:
                    initPars['delta_E'] = 0.
                    finalPars['delta_E'] = 0.

                initTune = np.sqrt(
                    (vInit * np.abs(initPars['eta_0']) *
                     np.sqrt(1 - (initPars['delta_E'] / vInit)**2)) /
                    (initPars['beta']**2 * initPars['energy']))

                finalTune = np.sqrt(
                    (vFin * np.abs(finalPars['eta_0']) *
                     np.sqrt(1 - (finalPars['delta_E'] / vFin)**2)) /
                    (finalPars['beta']**2 * finalPars['energy']))

                tDur = function_list[i][1][0] - fullTime[-1]
                nSteps = int(tDur / resolution[i - 1])
                time = np.linspace(float(fullTime[-1]), float(function_list[i][1][0]),
                                   nSteps)
                tuneInterp = np.linspace(float(initTune), float(finalTune), nSteps)

                mergePars = Ring.parameters_at_time(time)

                if main_h is False:
                    mergePars['delta_E'] *= 0

                volts = np.sqrt(
                    ((tuneInterp**2 * mergePars['beta']**2 *
                      mergePars['energy']) / (np.abs(mergePars['eta_0'])))**2 +
                    mergePars['delta_E']**2)

                fullFunction += volts.tolist() + 2 * [function_list[i][0]]
                fullTime += time.tolist() + function_list[i][1]

            else:

                start = np.where(function_list[i][0][0] >=
                                 function_list[i][1][0])[0][0]
                stop = np.where(function_list[i][0][0] >=
                                function_list[i][1][1])[0][0]

                funcTime = [function_list[i][1][0]] + \
                    function_list[i][0][0][start:stop].tolist() + \
                    [function_list[i][1][1]]
                funcProg = np.interp(funcTime, function_list[i][0][0],
                                     function_list[i][0][1])

                tDur = funcTime[0] - fullTime[-1]
                nSteps = int(tDur / resolution[i - 1])
                time = np.linspace(float(fullTime[-1]), float(funcTime[0]), nSteps)

                initPars = Ring.parameters_at_time(fullTime[-1])
                finalPars = Ring.parameters_at_time(funcTime[0])

                if main_h is False:
                    initPars['delta_E'] = 0.
                    finalPars['delta_E'] = 0.

                vInit = fullFunction[-1]
                vFin = funcProg[0]

                initTune = np.sqrt(
                    (vInit * np.abs(initPars['eta_0']) *
                     np.sqrt(1 - (initPars['delta_E'] / vInit)**2)) /
                    (initPars['beta']**2 * initPars['energy']))

                finalTune = np.sqrt(
                    (vFin * np.abs(finalPars['eta_0']) *
                     np.sqrt(1 - (finalPars['delta_E'] / vFin)**2)) /
                    (finalPars['beta']**2 * finalPars['energy']))

                tuneInterp = np.linspace(float(initTune), float(finalTune), nSteps)

                mergePars = Ring.parameters_at_time(time)

                if main_h is False:
                    mergePars['delta_E'] *= 0

                volts = np.sqrt(
                    ((tuneInterp**2 * mergePars['beta']**2 *
                      mergePars['energy']) / (np.abs(mergePars['eta_0'])))**2 +
                    mergePars['delta_E']**2)

                fullFunction += volts.tolist() + funcProg.tolist()
                fullTime += time.tolist() + funcTime

        else:
            # InputDataError
            raise RuntimeError("ERROR: merge_type not recognised")

    returnFunction = np.zeros([2, len(fullTime)])
    returnFunction[0] = fullTime
    returnFunction[1] = fullFunction

    return returnFunction
